module AI where

import Control.Concurrent.MVar

import State
import Dbg

---------------
-- COMPUTATIONS
---------------
--
-- computations run in threads and are killed after some time limit.
-- they communicate via MVars: the current state and the computation
-- result are read from a "result" MVar after the thread is killed.
-- the InitAI call is made once, and relevant StepAI calls are
-- repeated as the gameplay requires.
-- 
-- the InitAI call receives an empty state MVar, and each subsequent
-- StepAI call receives the state computed by from the last call.
-- 
-- each StepAI call receives an empty result MVar, which will be read after
-- the thread is killed.  if, at that point, the result MVar is empty,
-- the caller terminates with an error, so it is wise to seed the result var
-- early with a legal result.

type InitAI plugins_t state_t =
    Dbg -> plugins_t -> WorldSkeleton ->
    MVar (Maybe (plugins_t, state_t)) -> IO ()
type StepAI plugins_t state_t update_t result_t =
    Dbg -> plugins_t -> state_t -> update_t ->
    MVar (Maybe (plugins_t, state_t, result_t)) -> IO ()

set_state ::
    MVar (Maybe (plugins_t, state_t)) ->
    plugins_t -> state_t -> IO ()
set_state state_var plugins state =
    do _ <- swapMVar state_var (Just (plugins, state))
       return ()

set_result ::
    MVar (Maybe (plugins_t, state_t, result_t)) ->
    plugins_t -> state_t -> result_t -> IO ()
set_result result_var plugins state result =
    do _ <- swapMVar result_var (Just (plugins, state, result))
       return ()
