module CopMoveUtil where

import List
import Maybe

import CopPlugins
import State
import CopPluginLog
import LegalMoves
import RandomUtil

get_cops_to_move :: CopPlugins -> [PlayerName]
get_cops_to_move plugins =
    let ws = cpi_log_skeleton (cpi_log plugins)
        wu = cpi_log_update (cpi_log plugins)
        my_name = ws_my_name ws
        slave_cops = wu_slave_cops wu
        my_slave_cops = [slave |
                         (master, slave) <- slave_cops,
                         master == my_name]
        cops_to_move = (my_name:my_slave_cops)
    in  cops_to_move

state_to_player_move :: PlayerName -> PlayerState -> PlayerMove
state_to_player_move player_name player_state =
    PlayerMove { pm_name = player_name,
                 pm_state = player_state }

states_to_player_moves :: [PlayerName] -> [PlayerState] -> [PlayerMove]
states_to_player_moves cops cop_states =
    zipWith state_to_player_move cops cop_states

player_moves_to_cop_move :: [PlayerMove] -> CopMove
player_moves_to_cop_move player_moves =
    CopMove { cm_is_honest = True,
              cm_accusations = [],
              cm_moves = player_moves }

states_to_cop_move :: [PlayerName] -> [PlayerState] -> CopMove
states_to_cop_move cops cop_states =
    let player_moves = states_to_player_moves cops cop_states
        cop_move = player_moves_to_cop_move player_moves
    in  cop_move

get_identity_player_move :: CopPlugins -> PlayerName -> PlayerMove
get_identity_player_move plugins cop_name =
    let ws = cpi_log_skeleton (cpi_log plugins)
        wu = cpi_log_update (cpi_log plugins)
        idx = fromJust (elemIndex cop_name (ws_cops ws))
        current_state = (wu_cop_states wu) !! idx
    in  state_to_player_move cop_name current_state

get_identity_cop_move :: CopPlugins -> (CopMove, String)
get_identity_cop_move plugins =
    let cops_to_move = get_cops_to_move plugins
        player_moves = map (get_identity_player_move plugins) cops_to_move
        move = player_moves_to_cop_move player_moves
    in  (move, "identity")

get_random_player_move :: CopPlugins -> PlayerName -> IO PlayerMove
get_random_player_move plugins cop_name =
    do let ws = cpi_log_skeleton (cpi_log plugins)
           wu = cpi_log_update (cpi_log plugins)
           idx = fromJust (elemIndex cop_name (ws_cops ws))
           current_state = (wu_cop_states wu) !! idx
       random_state <- random_legal_move ws current_state
       return (state_to_player_move cop_name random_state)

get_random_cop_move :: CopPlugins -> IO (CopMove, String)
get_random_cop_move plugins =
    do let cops_to_move = get_cops_to_move plugins
       player_moves <- mapM (get_random_player_move plugins) cops_to_move
       let move = player_moves_to_cop_move player_moves
       return (move, "random")

get_random_non_identity_player_move :: CopPlugins -> PlayerName ->
                                       IO PlayerMove
get_random_non_identity_player_move plugins cop_name =
    do let ws = cpi_log_skeleton (cpi_log plugins)
           wu = cpi_log_update (cpi_log plugins)
           idx = fromJust (elemIndex cop_name (ws_cops ws))
           current_state = (wu_cop_states wu) !! idx
       random_state <- random_non_identity_legal_move ws current_state
       return (state_to_player_move cop_name random_state)

get_random_non_identity_cop_move :: CopPlugins -> IO (CopMove, String)
get_random_non_identity_cop_move plugins =
    do let cops_to_move = get_cops_to_move plugins
       player_moves <- mapM (get_random_non_identity_player_move plugins)
                       cops_to_move
       let move = player_moves_to_cop_move player_moves
       return (move, "random non-identity")

get_move_from_plans :: CopPlugins -> [Plan] -> PlayerName -> Maybe PlayerState
get_move_from_plans plugins plans cop_name =
    let ws = cpi_log_skeleton (cpi_log plugins)
        wu = cpi_log_update (cpi_log plugins)
        epoch = wu_epoch wu
        Just idx = elemIndex cop_name (ws_cops ws)
        current_ps = (wu_cop_states wu) !! idx
        all_moves = legal_moves ws current_ps
        planned_moves =
            [pm_state (pl_move plan) |
             plan <- plans,
             pm_name (pl_move plan) == cop_name,
             pl_epoch plan == epoch + 1,
             pm_state (pl_move plan) `elem` all_moves]
        planned_move =
            case planned_moves of
              [] -> Nothing
              [move] -> Just move
              _ -> Nothing
    in  planned_move

get_planned_player_move :: CopPlugins -> PlayerName -> PlayerMove
get_planned_player_move plugins cop_name =
    let sent_plans = cpi_log_sent_plans (cpi_log plugins)
        player_move =
            fromJust (get_move_from_plans plugins sent_plans cop_name)
    in  state_to_player_move cop_name player_move

-- Warning: get_planned_cop_move requires that we sent a planned move
get_planned_cop_move :: CopPlugins -> (CopMove, String)
get_planned_cop_move plugins =
    let cops_to_move = get_cops_to_move plugins
        player_moves = map (get_planned_player_move plugins) cops_to_move
        move = player_moves_to_cop_move player_moves
    in  (move, "planned")

get_voted_player_move :: CopPlugins -> PlayerName -> Maybe PlayerMove
get_voted_player_move plugins cop_name =
    let winner = cpi_log_winner (cpi_log plugins)
    in  case winner of
          Nothing -> Nothing
          Just player ->
              let received_plans = cpi_log_received_plans (cpi_log plugins)
                  maybe_plans = fromJust $ lookup player received_plans
                  plans = [fromJust maybe_plan
                           |
                           maybe_plan <- maybe_plans,
                           isJust maybe_plan]
                  maybe_move = get_move_from_plans plugins plans cop_name
              in  case maybe_move of
                    Nothing -> Nothing
                    Just move -> Just (state_to_player_move cop_name move)

get_voted_cop_move :: CopPlugins -> Maybe (CopMove, String)
get_voted_cop_move plugins =
    let cops_to_move = get_cops_to_move plugins
        maybe_player_moves = map (get_voted_player_move plugins) cops_to_move
    in  if all isJust maybe_player_moves
           then let player_moves = map fromJust maybe_player_moves
                    move = player_moves_to_cop_move player_moves
                in  Just (move, "planned")
           else Nothing

{-
get_voted_move :: CopPlugins -> Maybe (PlayerState, String)
get_voted_move plugins =
    let winner = cpi_log_winner (cpi_log plugins)
    in  case winner of
          Nothing -> Nothing
          Just player ->
              let ws = cpi_log_skeleton (cpi_log plugins)
                  cops = ws_cops ws
                  Just player_idx = elemIndex player cops
                  received_plans = cpi_log_received_plans (cpi_log plugins)
                  maybe_plans = received_plans !! player_idx
                  plans = [fromJust maybe_plan
                           |
                           maybe_plan <- maybe_plans,
                           isJust maybe_plan]
                  maybe_move = get_move_from_plans plugins plans
              in  case maybe_move of
                    Nothing -> Nothing
                    Just move -> Just (move, "voted")
-}
