module CopPluginTracking where

import Data.Array
import qualified Data.Set as S
import qualified Data.Map as M
import Maybe

import State
import Tracking

data CopPluginTracking = CopPluginTracking {
    cpi_tracking_ws :: WorldSkeleton,
    cpi_tracking_wu :: WorldUpdate,
    cpi_tracking_all :: M.Map [Bool] Tracking,
    cpi_tracking_other_cops :: [PlayerName],
    cpi_tracking_non_liars :: [Bool],
    cpi_tracking_best_key :: [Bool],
    cpi_tracking_evidence_pre_update :: EvidenceTracking,
    cpi_tracking_evidence :: EvidenceTracking,
    cpi_tracking_evidence_senders :: [Maybe Bool],
    cpi_tracking_use_sketchy_evidence :: Bool,
    cpi_tracking_echo_after :: Maybe Epoch
}

-- How to use the robber and evidence tracking:
--
-- cpi_tracking_get_best_states :: CopPluginTracking -> PlayerStateSet
-- 
--   cpi_tracking_get_best_states gives you all the states where the
--   robbber might be now.
--
-- type EvidenceTrackingItem = (Epoch, PlayerStateSet)
-- type EvidenceTracking = [EvidenceTrackingItem]
-- cpi_tracking_get_best_evidence :: CopPluginTracking -> EvidenceTracking
--
--   cpi_tracking_get_best_evidence give information about all pieces
--   of evidence that might be present now: both what epoch the
--   evidence was dropped and where it might be.  They are returned in
--   chronologically increasing order.

cpi_tracking_get_best_states :: CopPluginTracking -> PlayerStateSet
cpi_tracking_get_best_states plugin =
    let key = cpi_tracking_best_key plugin
    in  cpi_tracking_get_states plugin key

cpi_tracking_get_best_tracking :: CopPluginTracking -> Tracking
cpi_tracking_get_best_tracking plugin =
    let key = cpi_tracking_best_key plugin
    in  cpi_tracking_get_tracking plugin key

cpi_tracking_get_best_evidence :: CopPluginTracking -> EvidenceTracking
cpi_tracking_get_best_evidence plugin =
    let key = cpi_tracking_best_key plugin
    in  cpi_tracking_get_evidence plugin key

cpi_tracking_get_my_states :: CopPluginTracking -> PlayerStateSet
cpi_tracking_get_my_states plugin =
    let key = cpi_tracking_get_my_key plugin
    in  cpi_tracking_get_states plugin key

cpi_tracking_get_my_tracking :: CopPluginTracking -> Tracking
cpi_tracking_get_my_tracking plugin =
    let key = cpi_tracking_get_my_key plugin
    in  cpi_tracking_get_tracking plugin key

cpi_tracking_get_my_evidence :: CopPluginTracking -> EvidenceTracking
cpi_tracking_get_my_evidence plugin =
    let key = cpi_tracking_get_my_key plugin
    in  cpi_tracking_get_evidence plugin key

cpi_tracking_get_states :: CopPluginTracking -> [Bool] -> PlayerStateSet
cpi_tracking_get_states plugin key =
    let tracking = cpi_tracking_get_tracking plugin key
        epoch = wu_epoch (cpi_tracking_wu plugin)
        idx = tracking_epoch_to_idx epoch
    in  tracking ! idx

cpi_tracking_get_tracking :: CopPluginTracking -> [Bool] -> Tracking
cpi_tracking_get_tracking plugin key =
    let Just tracking = M.lookup key (cpi_tracking_all plugin)
    in  tracking

cpi_tracking_get_evidence :: CopPluginTracking -> [Bool] -> EvidenceTracking
cpi_tracking_get_evidence plugin key =
    let tracking = cpi_tracking_get_tracking plugin key
        et = cpi_tracking_evidence plugin
        et' = evidence_tracking_apply_robber_tracking et tracking
    in  et'

cpi_tracking_get_informs :: CopPluginTracking -> [Inform]
cpi_tracking_get_informs plugin =
    let ws = cpi_tracking_ws plugin
        wu = cpi_tracking_wu plugin
        epoch = wu_epoch wu
        idx = tracking_epoch_to_idx epoch
        echo_after = cpi_tracking_echo_after plugin
        key = if isJust echo_after && epoch > fromJust echo_after
                 then cpi_tracking_best_key plugin
                 else cpi_tracking_get_my_key plugin
        Just tracking = M.lookup key (cpi_tracking_all plugin)
        possible = tracking ! idx
        all_states = (map (loc_to_state PT_robber) (ws_all_locs ws))
        impossible = (S.fromList all_states) `S.difference` possible
    in  -- All the places we know the robber is not get certainty -100.
        [Inform { if_who = ws_robber ws,
                  if_loc = ps_loc state,
                  if_ptype = ps_type state,
                  if_epoch = epoch,
                  if_certainty = -100 }
         |
         state <- S.toList impossible]
        ++
        -- If we know exactly where the robber is, that location gets
        -- certainty 100.  If there are multiple possible locations,
        -- no certainty value is assigned to any of them.
        if S.size possible == 1
        then [Inform { if_who = ws_robber ws,
                       if_loc = ps_loc state,
                       if_ptype = ps_type state,
                       if_epoch = epoch,
                       if_certainty = 100 }
              |
              state <- S.toList possible]
        else []
        ++
        -- For each piece of evidence visible, the evidence location
        -- and time get certainty 100.
        [Inform { if_who = ws_robber ws,
                  if_loc = ev_loc,
                  if_ptype = PT_robber,
                  if_epoch = ev_epoch,
                  if_certainty = 100 }
         |
         (ev_loc, ev_epoch) <- wu_evidence wu]

cpi_tracking_get_my_key :: CopPluginTracking -> [Bool]
cpi_tracking_get_my_key plugin =
    map (const False) (cpi_tracking_other_cops plugin)

cpi_tracking_get_one_cop_key :: CopPluginTracking -> Int -> [Bool]
cpi_tracking_get_one_cop_key plugin cop_num =
    map (== cop_num) [0..(length (cpi_tracking_other_cops plugin) - 1)]
