module CopPluginTrackingImpl where

import Data.Maybe
import qualified Data.Set as S
import List
import qualified Data.Map as M

import State
import CopPluginTracking
import Tracking
import CopPlugins
import LegalMoves
import CopPluginCache

cpi_tracking_init :: CopPlugins -> WorldSkeleton -> CopPluginTracking
cpi_tracking_init _ ws =
    let other_cops = delete (ws_my_name ws) (ws_cops ws)
        num_other_cops = length other_cops
        all_keys = get_all_keys num_other_cops
        initial_tracking = tracking_init ws
    in  CopPluginTracking {
            cpi_tracking_ws = ws,
            cpi_tracking_wu = error "No update",
            cpi_tracking_all =
                M.fromList [(k, initial_tracking) | k <- all_keys],
            cpi_tracking_other_cops = other_cops,
            cpi_tracking_non_liars = map (const True) other_cops,
            cpi_tracking_best_key = map (const True) other_cops,
            cpi_tracking_evidence_pre_update = evidence_tracking_init,
            cpi_tracking_evidence = evidence_tracking_init,
            cpi_tracking_evidence_senders = map (const Nothing) other_cops,
            cpi_tracking_use_sketchy_evidence = False,
            cpi_tracking_echo_after = Just 100
        }

get_all_keys :: Int -> [[Bool]]
get_all_keys 0 = [[]]
get_all_keys n = [(b:keys) | b <- [False, True], keys <- get_all_keys (n-1)]

cpi_tracking_pre_inform ::
    CopPlugins -> CopPlugins -> CopPluginTracking -> WorldUpdate ->
    CopPluginTracking
cpi_tracking_pre_inform _ plugins plugin wu =
    let ws = cpi_tracking_ws plugin
        tracking_all = cpi_tracking_all plugin
        tracking_all' = fmap (cop_tracking_update plugins ws wu et')
                        tracking_all
        et = cpi_tracking_evidence plugin
        et' = evidence_tracking_advance ((wu_epoch wu) - 1) et
        et'' = evidence_tracking_update ws wu et'
        plugin' = plugin { cpi_tracking_wu = wu,
                           cpi_tracking_all = tracking_all',
                           cpi_tracking_evidence_pre_update = et',
                           cpi_tracking_evidence = et'' }
    in  catch_liars plugin'

cop_tracking_update :: CopPlugins ->
                       WorldSkeleton -> WorldUpdate -> EvidenceTracking ->
                       Tracking -> Tracking
cop_tracking_update plugins ws wu et tracking0 =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        tracking1 = tracking_advance legal_moves_cached tracking0
        tracking2 = cop_tracking_update_cops plugins ws wu tracking1
        tracking3 = cop_tracking_update_banks plugins ws wu tracking2
        tracking4 = cop_tracking_update_smell plugins ws wu tracking3
        tracking5 = cop_tracking_update_evidence plugins ws wu tracking4
        tracking6 = cop_tracking_update_missing_evidence plugins
                    ws wu et tracking5
        tracking' = tracking6
    in  tracking'

cop_tracking_update_cops ::
    CopPlugins -> WorldSkeleton -> WorldUpdate -> Tracking -> Tracking
cop_tracking_update_cops plugins ws wu tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
        -- Robber must not be anywhere a cop is
        now = wu_epoch wu
        cop_states = wu_cop_states wu
        impossible_states =
            map ((loc_to_state PT_robber) . state_to_loc) cop_states
        impossible_states_set = S.fromList impossible_states
    in  tracking_update_impossible legal_moves_cached legal_anti_moves_cached
        now impossible_states_set tracking

cop_tracking_update_banks ::
    CopPlugins -> WorldSkeleton -> WorldUpdate -> Tracking -> Tracking
cop_tracking_update_banks plugins ws wu tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
        now = wu_epoch wu
    in  case wu_robber_state wu of
            Just robber_state ->
                -- Robber is robbing a bank, so he must be at that bank
                tracking_update_possible
                legal_moves_cached
                legal_anti_moves_cached
                now
                (S.singleton robber_state)
                tracking
            Nothing ->
                -- Robber is not robbing a bank, so he must not be at any bank
                let bank_states = map (loc_to_state PT_robber) (ws_banks ws)
                in  tracking_update_impossible
                    legal_moves_cached
                    legal_anti_moves_cached
                    now
                    (S.fromList bank_states)
                    tracking

cop_tracking_update_smell ::
    CopPlugins -> WorldSkeleton -> WorldUpdate -> Tracking -> Tracking
cop_tracking_update_smell plugins ws wu tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
        now = wu_epoch wu
        my_name = ws_my_name ws
        Just my_idx = elemIndex my_name (ws_cops ws)
        my_state = (wu_cop_states wu) !! my_idx
        (cop_states_up_to, cop_states_at) =
            states_by_distance_same_state ws my_state
        robbify_set states =
            S.fromList $
            map (((loc_to_state PT_robber) . state_to_loc)) (S.toList states)
        states_at = map robbify_set cop_states_at
    in  case wu_smell wu of
            Just dist ->
                -- Cop can smell the robber, so he must be at the smell
                -- distance
                tracking_update_possible
                legal_moves_cached
                legal_anti_moves_cached
                now (states_at !! dist) tracking
            Nothing ->
                -- Cop cannot smell the robber, so he must not be within
                -- the distance that the cop can smell
                case ps_type my_state of
                    PT_cop_car ->
                        tracking_update_impossible
                        legal_moves_cached
                        legal_anti_moves_cached
                        now (states_at !! 1) tracking
                    PT_cop_foot ->
                        tracking_update_impossible
                        legal_moves_cached
                        legal_anti_moves_cached
                        now
                        ((states_at !! 1) `S.union` (states_at !! 2))
                        tracking

cop_tracking_update_evidence ::
    CopPlugins ->
    WorldSkeleton -> WorldUpdate -> Tracking -> Tracking
cop_tracking_update_evidence plugins ws wu tracking =
    let evs = wu_evidence wu
    in  cop_tracking_update_evidence' plugins ws evs tracking

cop_tracking_update_evidence' _ _ [] tracking = tracking
cop_tracking_update_evidence' plugins ws (ev:evs) tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
        (loc, epoch) = ev
    in  tracking_update_possible
        legal_moves_cached
        legal_anti_moves_cached
        epoch
        (S.singleton (loc_to_state PT_robber loc)) tracking

cop_tracking_update_missing_evidence ::
    CopPlugins ->
    WorldSkeleton -> WorldUpdate -> EvidenceTracking -> Tracking -> Tracking
cop_tracking_update_missing_evidence plugins ws wu [] tracking = tracking
cop_tracking_update_missing_evidence plugins ws wu (item:items) tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
        (epoch, impossible_states) = item
        evs = wu_evidence wu
        my_name = ws_my_name ws
        Just my_idx = elemIndex my_name (ws_cops ws)
        current_ps = (wu_cop_states wu) !! my_idx
        loc = state_to_loc current_ps
        robber_ps = loc_to_state PT_robber loc
        tracking' =
            if robber_ps `S.member` impossible_states
               then tracking
               else if (loc, epoch) `elem` evs
                       then tracking
                            {-
                            tracking_update_possible
                            legal_moves_cached
                            legal_anti_moves_cached
                            epoch
                            (S.singleton (loc_to_state PT_robber loc)) tracking
                            -}
                       else tracking_update_impossible
                            legal_moves_cached
                            legal_anti_moves_cached
                            epoch
                            (S.singleton (loc_to_state PT_robber loc)) tracking
    in  cop_tracking_update_missing_evidence plugins ws wu items tracking'

cpi_tracking_pre_plan ::
    CopPlugins -> CopPlugins -> CopPluginTracking ->
    [(PlayerName, [Maybe Inform])] ->
    CopPluginTracking
cpi_tracking_pre_plan _ plugins plugin informs =
    let ws = cpi_tracking_ws plugin
        wu = cpi_tracking_wu plugin
        other_informs = map informs_for_cop (cpi_tracking_other_cops plugin)
        et_pre_update = cpi_tracking_evidence_pre_update plugin
        evidence_senders = cpi_tracking_evidence_senders plugin
        evidence_senders' =
            zipWith (update_evidence_sender ws wu et_pre_update)
                    evidence_senders other_informs
        filtered_informs = map (filter_informs ws wu) other_informs
        informs_by_epoch = map sift_by_epoch filtered_informs
        tracking_all = cpi_tracking_all plugin
        tracking_all' =
            update_tracking_all plugins plugin informs_by_epoch
                                evidence_senders'
                                tracking_all
        plugin' = plugin { cpi_tracking_all = tracking_all',
                           cpi_tracking_evidence_senders =
                               evidence_senders' }
    in  catch_liars plugin'
    where
    informs_for_cop cop_name =
        case lookup cop_name informs of
          Just one_cop_informs -> one_cop_informs
          Nothing -> []

filter_informs :: WorldSkeleton -> WorldUpdate -> [Maybe Inform] -> [Inform]
filter_informs ws wu informs =
    [inform |
     maybe_inform <- informs,
     isJust maybe_inform,
     let inform = fromJust maybe_inform,
     if_who inform == ws_robber ws,
     if_ptype inform == PT_robber,
     if_epoch inform >= 0 && if_epoch inform <= wu_epoch wu,
     if_certainty inform == -100 || if_certainty inform == 100]

update_evidence_sender :: WorldSkeleton -> WorldUpdate -> EvidenceTracking ->
                          Maybe Bool -> [Maybe Inform] -> Maybe Bool
update_evidence_sender _ _ et (Just False) _ = Just False
update_evidence_sender ws wu et old_val informs =
    let evidence_informs = filter_evidence_informs ws wu et informs
        normal_evidence_informs =
            filter_normal_evidence_informs ws wu et informs
    in  if length normal_evidence_informs < length evidence_informs
           then Just False
           else if length normal_evidence_informs > 0
                   then Just True
                   else old_val

filter_evidence_informs ::
    WorldSkeleton -> WorldUpdate -> EvidenceTracking ->
    [Maybe Inform] -> [Inform]
filter_evidence_informs ws wu et informs =
    [inform |
     maybe_inform <- informs,
     isJust maybe_inform,
     let inform = fromJust maybe_inform,
     if_epoch inform < wu_epoch wu]

filter_normal_evidence_informs ::
    WorldSkeleton -> WorldUpdate -> EvidenceTracking ->
    [Maybe Inform] -> [Inform]
filter_normal_evidence_informs ws wu et informs =
    [inform |
     maybe_inform <- informs,
     isJust maybe_inform,
     let inform = fromJust maybe_inform,
     if_who inform == ws_robber ws,
     if_ptype inform == PT_robber,
     if_epoch inform >= 0 && if_epoch inform < wu_epoch wu,
     isJust (lookup (if_epoch inform) et),
     if_certainty inform == 100]

sift_by_epoch :: [Inform] -> M.Map Epoch (PlayerStateSet, PlayerStateSet)
sift_by_epoch [] = M.empty
sift_by_epoch (inform:rest) =
    let rest_map = sift_by_epoch rest
        epoch = if_epoch inform
        ps = PlayerState { ps_loc = if_loc inform,
                           ps_type = if_ptype inform }
        possible = (if_certainty inform == 100)
        ((cur_possible, cur_impossible), rest_map') =
            case M.lookup epoch rest_map of
              Nothing -> ((S.empty, S.empty), rest_map)
              Just cur_entry -> (cur_entry, M.delete epoch rest_map)
        new_entry =
            if possible
              then (S.insert ps cur_possible, cur_impossible)
              else (cur_possible, S.insert ps cur_impossible)
        new_map = M.insert epoch new_entry rest_map'
    in  new_map

update_tracking_all ::
    CopPlugins -> CopPluginTracking ->
    [M.Map Epoch (PlayerStateSet, PlayerStateSet)] -> [Maybe Bool] ->
    M.Map [Bool] Tracking -> M.Map [Bool] Tracking
update_tracking_all plugins plugin informs_by_epoch evidence_senders
                    tracking_all =
    let ws = cpi_tracking_ws plugin
        other_cops = cpi_tracking_other_cops plugin
        num_other_cops = length other_cops
        all_keys = get_all_keys num_other_cops
        evidence_senders_to_use =
            if cpi_tracking_use_sketchy_evidence plugin
               then evidence_senders
               else map (const (Just False)) evidence_senders
        et_pre_update = cpi_tracking_evidence_pre_update plugin
        tracking_all' =
            M.fromList [(k,
                         update_tracking_key k informs_by_epoch
                         evidence_senders_to_use other_cops et_pre_update
                         (fromJust (M.lookup k tracking_all)))
                        |
                        k <- all_keys]
    in  tracking_all'
    where
    update_tracking_key [] [] [] [] _ tracking = tracking
    update_tracking_key (False:ks) (_:is) (_:ess) (_:cs) etpu tracking =
        update_tracking_key ks is ess cs etpu tracking
    update_tracking_key (True:ks) (i:is) (es:ess) (c:cs) etpu tracking =
        apply_informs_to_tracking
        plugins
        (cpi_tracking_ws plugin)
        (cpi_tracking_wu plugin)
        i
        c
        es
        etpu
        (update_tracking_key ks is ess cs etpu tracking)

apply_informs_to_tracking ::
    CopPlugins ->
    WorldSkeleton -> WorldUpdate ->
    M.Map Epoch (PlayerStateSet, PlayerStateSet) ->
    PlayerName ->
    Maybe Bool -> EvidenceTracking -> Tracking -> Tracking
apply_informs_to_tracking plugins ws wu informs_by_epoch
                          cop_name sends_evidence et_pre_update
                          tracking =
    let epochs = M.assocs informs_by_epoch
        apply_epoch_to_tracking t (epoch, pos_impos) =
            apply_one_epoch_informs_to_tracking plugins ws epoch pos_impos t
        tracking' = foldl apply_epoch_to_tracking tracking epochs
        tracking'' = if (sends_evidence == Just True)
                        then apply_sketchy_evidence plugins ws wu
                             informs_by_epoch cop_name et_pre_update tracking'
                        else tracking'
    in  tracking''

apply_one_epoch_informs_to_tracking ::
    CopPlugins ->
    WorldSkeleton -> Epoch -> (PlayerStateSet, PlayerStateSet) -> Tracking ->
    Tracking
apply_one_epoch_informs_to_tracking plugins ws epoch
                                    (possible, impossible) tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
    in  case S.size possible of
          0 -> case S.size impossible of
                 0 -> tracking
                 _ -> tracking_update_impossible
                      legal_moves_cached
                      legal_anti_moves_cached
                      epoch impossible tracking
          1 -> tracking_update_possible
               legal_moves_cached
               legal_anti_moves_cached
               epoch possible tracking
          _ -> tracking_update_possible
               legal_moves_cached
               legal_anti_moves_cached
               epoch (S.empty) tracking

apply_sketchy_evidence :: CopPlugins ->
                          WorldSkeleton -> WorldUpdate ->
                          M.Map Epoch (PlayerStateSet, PlayerStateSet) ->
                          PlayerName -> EvidenceTracking -> Tracking ->
                          Tracking
apply_sketchy_evidence _ _ _ _ _ [] tracking = tracking
apply_sketchy_evidence plugins ws wu informs_by_epoch cop_name
                       (item:items) tracking =
    let legal_moves_cached =
            cpi_cache_legal_moves (cpi_cache plugins)
        legal_anti_moves_cached =
            cpi_cache_legal_anti_moves (cpi_cache plugins)
        (epoch, impossible_states) = item
        Just cop_idx = elemIndex cop_name (ws_cops ws)
        ps = (wu_cop_states wu) !! cop_idx
        loc = state_to_loc ps
        robber_ps = loc_to_state PT_robber loc
        tracking' =
            if robber_ps `S.member` impossible_states
               then tracking
               else case M.lookup epoch informs_by_epoch of
                      Just _ -> tracking
                      Nothing -> tracking_update_impossible
                                 legal_moves_cached
                                 legal_anti_moves_cached
                                 epoch
                                 (S.singleton (loc_to_state PT_robber loc))
                                 tracking
    in  apply_sketchy_evidence
        plugins ws wu informs_by_epoch cop_name items tracking'

catch_liars :: CopPluginTracking -> CopPluginTracking
catch_liars plugin =
    let states = cpi_tracking_get_best_states plugin
        plugin' =
            if S.null states
              then let other_cops = cpi_tracking_other_cops plugin
                       num_other_cops = length other_cops
                       current_non_liars =
                           map (is_non_liar plugin) [0..(num_other_cops - 1)]
                       new_non_liars = zipWith (&&)
                                       (cpi_tracking_non_liars plugin)
                                       current_non_liars
                   in  plugin { cpi_tracking_non_liars = new_non_liars }
              else plugin
        states' = cpi_tracking_get_best_states plugin'
        new_best_key =
            if S.null states'
              then cpi_tracking_get_my_key plugin
              else cpi_tracking_non_liars plugin
    in plugin' { cpi_tracking_best_key = new_best_key }

is_non_liar :: CopPluginTracking -> Int -> Bool
is_non_liar plugin cop_num =
    let key = cpi_tracking_get_one_cop_key plugin cop_num
        states = cpi_tracking_get_states plugin key
    in  not (S.null states)
