module LegalMoves where

import Data.Graph.Inductive
import Maybe
import qualified Data.Set as S
import List

import State
import RandomUtil
import CacheUtil

legal_moves :: WorldSkeleton -> PlayerState -> [PlayerState]
legal_moves ws ps =
    let m = ws_map_ps ws
    in  (ps:[fromJust (lab m n) | n <- (suc m ((ws_ps_to_node ws) ps))])

legal_moves_same_state :: WorldSkeleton -> PlayerState -> [PlayerState]
legal_moves_same_state ws ps@(PlayerState { ps_type = pt }) =
    [move | move <- legal_moves ws ps, ps_type move == pt]

legal_anti_moves :: WorldSkeleton -> PlayerState -> [PlayerState]
legal_anti_moves ws ps =
    let m = ws_map_ps ws
    in  (ps:[fromJust (lab m n) | n <- (pre m ((ws_ps_to_node ws) ps))])

type LegalMovesFn = PlayerState -> [PlayerState]

make_legal_moves_cache_fn :: WorldSkeleton -> LegalMovesFn -> LegalMovesFn
make_legal_moves_cache_fn ws next_moves =
    let all_states = map snd (labNodes (ws_map_ps ws))
    in  make_cache_fn next_moves all_states

type StatesUpToDistance = [PlayerStateSet]
type StatesAtDistance = [PlayerStateSet]
type LocsUpToDistance = [LocSet]
type LocsAtDistance = [LocSet]

type StatesUpToDistanceFn = PlayerState -> StatesUpToDistance
type StatesAtDistanceFn = PlayerState -> StatesAtDistance
type LocsUpToDistanceFn = PlayerState -> LocsUpToDistance
type LocsAtDistanceFn = PlayerState -> LocsAtDistance

states_by_distance ::
    WorldSkeleton -> PlayerState -> (StatesUpToDistance, StatesAtDistance)
states_by_distance ws ps =
    states_by_distance_generic (legal_moves ws) ps

states_by_distance_same_state ::
    WorldSkeleton -> PlayerState -> (StatesUpToDistance, StatesAtDistance)
states_by_distance_same_state ws ps =
    states_by_distance_generic (legal_moves_same_state ws) ps

states_by_distance_generic :: (PlayerState -> [PlayerState]) -> PlayerState ->
                              (StatesUpToDistance, StatesAtDistance)
states_by_distance_generic next_moves ps =
    let states_up_to_distance =
            ((S.singleton ps) :
             states_up_to_distance' states_at_distance states_up_to_distance)
        states_at_distance =
            ((S.singleton ps) :
             states_at_distance' states_up_to_distance)
        states_up_to_distance' (at_x:at_xs) (up_to_x:up_to_xs) =
            (((S.fromList (concatMap next_moves (S.toList at_x)))
              `S.union` up_to_x) :
             states_up_to_distance' at_xs up_to_xs)
        states_at_distance' (up_to_x:up_to_xs@(up_to_x2:up_to_xs2)) =
             (up_to_x2 `S.difference` up_to_x :
              states_at_distance' up_to_xs)
    in (states_up_to_distance, states_at_distance)

states_and_locs_by_distance_generic ::
    (PlayerState -> [PlayerState]) -> PlayerState ->
    ((StatesUpToDistance, StatesAtDistance),
     (LocsUpToDistance, LocsAtDistance))
states_and_locs_by_distance_generic next_moves ps =
    let (states_up_to_distance, states_at_distance) =
            states_by_distance_generic next_moves ps
        locs_up_to_distance =
            map (S.fromList . (map state_to_loc) . S.toList)
                states_up_to_distance
        locs_at_distance =
            zipWith S.difference
                    locs_up_to_distance
                    (S.empty:locs_up_to_distance)
    in ((states_up_to_distance, states_at_distance),
        (locs_up_to_distance, locs_at_distance))

make_by_distance_cache_fns ::
    WorldSkeleton -> (PlayerState -> [PlayerState]) ->
    ((StatesUpToDistanceFn, StatesAtDistanceFn),
     (LocsUpToDistanceFn, LocsAtDistanceFn))
make_by_distance_cache_fns ws next_moves =
    let all_states = map snd (labNodes (ws_map_ps ws))
        fn = make_cache_fn
             (states_and_locs_by_distance_generic next_moves)
             all_states
    in (((\ ps -> fst (fst (fn ps))),
         (\ ps -> snd (fst (fn ps)))),
        ((\ ps -> fst (snd (fn ps))),
         (\ ps -> snd (snd (fn ps)))))

get_states_at_distance ::
    WorldSkeleton -> PlayerState -> Int -> PlayerStateSet
get_states_at_distance ws ps dist =
    let (up_to_distance, at_distance) = states_by_distance ws ps
    in  at_distance !! dist

get_states_up_to_distance ::
    WorldSkeleton -> PlayerState -> Int -> PlayerStateSet
get_states_up_to_distance ws ps dist =
    let (up_to_distance, at_distance) = states_by_distance ws ps
    in  up_to_distance !! dist

random_legal_move :: WorldSkeleton -> PlayerState -> IO PlayerState
random_legal_move ws ps =
    do let legal_states = legal_moves ws ps
       random_ps <- random_list_elt legal_states
       return random_ps

random_non_identity_legal_move ::
    WorldSkeleton -> PlayerState -> IO PlayerState
random_non_identity_legal_move ws ps =
    do let legal_states = legal_moves ws ps
           legal_non_identity_states = delete ps legal_states
           states_to_use =
               if null legal_non_identity_states
                 then legal_states
                 else legal_non_identity_states
       random_ps <- random_list_elt states_to_use
       return random_ps
