module Minimax where

import List
import Maybe
import Random

import RandomUtil

type Depth = Int

data MinimaxOpts state_t move_t score_t = MinimaxOpts {
    mmo_moves :: Depth -> state_t -> [move_t],
    mmo_advance :: state_t -> move_t -> state_t,
    mmo_eval :: state_t -> score_t,
    mmo_want_max :: state_t -> Bool,
    mmo_next_depth :: Depth -> Depth,
    mmo_default_moves :: [Maybe move_t]
}

default_minimax_opts = MinimaxOpts {
    mmo_moves = error "mmo_moves undefined",
    mmo_advance = error "mmo_advance undefined",
    mmo_eval = error "mmo_eval undefined",
    mmo_want_max = error "mmo_want_max undefined",
    mmo_next_depth = (+1),
    mmo_default_moves = []
}

data Infinity a = NegInfinity | Finite a | Infinity
    deriving (Eq, Ord)

minimax :: (Ord score_t, Eq move_t, RandomGen randomgen_t) =>
           MinimaxOpts state_t move_t score_t ->
           randomgen_t ->
           state_t ->
           [(score_t, [move_t])]
minimax opts g state =
    let depths = drop 1 (iterate (mmo_next_depth opts) 0)
    in  minimax' opts g state depths

minimax' :: (Ord score_t, Eq move_t, RandomGen randomgen_t) =>
            MinimaxOpts state_t move_t score_t ->
            randomgen_t ->
            state_t ->
            [Depth] ->
            [(score_t, [move_t])]
minimax' opts g state (depth:depths) =
    let (g1, g2) = split g
    in  (minimax_1 opts g1 state depth:minimax' opts g2 state depths)

minimax_1 :: (Ord score_t, Eq move_t, RandomGen randomgen_t) =>
             MinimaxOpts state_t move_t score_t ->
             randomgen_t ->
             state_t ->
             Depth ->
             (score_t, [move_t])
minimax_1 opts g state depth =
    let default_moves = mmo_default_moves opts
        (Finite score, moves) =
            minimax_1'
            opts
            g
            (mmo_default_moves opts)
            NegInfinity
            Infinity
            depth
            state
    in  (score, moves)

minimax_1' :: (Ord score_t, Eq move_t, RandomGen randomgen_t) =>
              MinimaxOpts state_t move_t score_t ->
              randomgen_t ->
              [Maybe move_t] ->
              Infinity score_t ->
              Infinity score_t ->
              Depth ->
              state_t ->
              (Infinity score_t, [move_t])
minimax_1' opts g default_moves alpha beta depth state =
    if depth == 0
      then (Finite (mmo_eval opts state), [])
      else let all_moves = mmo_moves opts depth state
           in  if null all_moves -- game over, in other words
                 then (Finite (mmo_eval opts state), [])
                 else let (randomized_moves, g') =
                              --random_list_perm_fn all_moves g
                              (all_moves, g)
                          default_plus_all_moves =
                              case default_moves of
                                (Just default_move:_) ->
                                    (default_move:
                                     delete default_move randomized_moves)
                                _ -> randomized_moves
                      in  minimax_1'' opts g default_moves alpha beta depth
                          state default_plus_all_moves

minimax_1'' :: (Ord score_t, Eq move_t, RandomGen randomgen_t) =>
               MinimaxOpts state_t move_t score_t ->
               randomgen_t ->
               [Maybe move_t] ->
               Infinity score_t ->
               Infinity score_t ->
               Depth ->
               state_t ->
               [move_t] ->
               (Infinity score_t, [move_t])
minimax_1'' opts g default_moves alpha beta depth state moves =
    let want_max = mmo_want_max opts state
        best_score = if want_max then alpha else beta
    in  minimax_1''' opts g default_moves alpha beta depth state
        (error "Minimax error") (best_score, error "Minimax error") moves

minimax_1''' :: (Ord score_t, Eq move_t, RandomGen randomgen_t) =>
                MinimaxOpts state_t move_t score_t ->
                randomgen_t ->
                [Maybe move_t] ->
                Infinity score_t ->
                Infinity score_t ->
                Depth ->
                state_t ->
                move_t ->
                (Infinity score_t, [move_t]) ->
                [move_t] ->
                (Infinity score_t, [move_t])
minimax_1''' opts g default_moves alpha beta depth state
            best_move (best_score, best_next_moves) [] =
    (best_score, (best_move:best_next_moves))
minimax_1''' opts g default_moves alpha beta depth state
            best_move (best_score, best_next_moves) (move:moves) =
    let (g1, g2) = split g
        want_max = mmo_want_max opts state
        next_state = mmo_advance opts state move
        (score, next_moves) =
            minimax_1' opts g1 (drop 1 default_moves) alpha beta
                       (depth-1) next_state
        (alpha', beta') =
            if want_max
              then (max alpha score, beta)
              else (alpha, min beta score)
    in  if beta' <= alpha'
          then (if want_max then beta else alpha, error "Minimax error")
          else let better = if want_max
                              then alpha' > alpha
                              else beta' < beta
                   (new_best_move, new_best_score, new_best_next_moves) =
                       if better
                         then (move, score, next_moves)
                         else (best_move, best_score, best_next_moves)
               in  minimax_1''' opts g2 default_moves alpha' beta' depth state
                   new_best_move (new_best_score, new_best_next_moves) moves
