module Parser where

import List
import qualified Data.Graph as G
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Token
import Data.Graph.Inductive
import qualified Data.Map as M
import Maybe

import State

-- XXX should keywords be reserved?

type NameToLoc = LocName -> Maybe Loc
type LocToName = Loc -> LocName

-- winner message --

p_winner :: Parser Winner
p_winner  =  (p_keyword "winner:" >> p_name >>= return . Just)
         <|> (p_keyword "nowinner:" >> return Nothing)

-- plan messages --

p_from_plans :: NameToLoc -> [PlayerName] ->
                Parser [(PlayerName, [Maybe Plan])]
p_from_plans name_to_loc cops =
    p_froms cops (p_plans name_to_loc)

p_plans :: NameToLoc -> Parser [Maybe Plan]
p_plans name_to_loc =
    do p_keyword "plan\\"
       plans <- many (p_plan name_to_loc)
       p_keyword "plan/"
       return plans

p_plan :: NameToLoc -> Parser (Maybe Plan)
p_plan name_to_loc =
    do p_keyword "plan:"
       who <- p_name <?> "plan protagonist"
       dest_name <- p_name <?> "plan destination"
       job <- p_player_type <?> "plan mode of transportation"
       epoch <- p_number <?> "plan epoch"
       case name_to_loc dest_name of
           Just dest ->
               return
               (Just
                (Plan { pl_epoch = epoch,
                        pl_move = PlayerMove {
                                      pm_state =
                                          PlayerState { ps_loc = dest,
                                                        ps_type = job },
                                      pm_name = who }}))
           Nothing -> return Nothing

-- inform messages --

p_from_informs :: NameToLoc -> [PlayerName] ->
                  Parser [(PlayerName, [Maybe Inform])]
p_from_informs name_to_loc cops =
    p_froms cops (p_informs name_to_loc)

p_informs :: NameToLoc -> Parser [Maybe Inform]
p_informs name_to_loc =
    do p_keyword "inf\\"
       informs <- many (p_inform name_to_loc)
       p_keyword "inf/"
       return informs

p_inform :: NameToLoc -> Parser (Maybe Inform)
p_inform name_to_loc =
    do p_keyword "inf:"
       who <- p_name <?> "player name"
       loc_name <- p_name <?> "player location"
       job <- p_player_type <?> "player type"
       epoch <- p_number <?> "epoch"
       certainty <- p_number <?> "certainty"
       case name_to_loc loc_name of
           Just loc ->
               return (Just
                       Inform { if_who = who, if_loc = loc,
                                if_ptype = job, if_epoch = epoch,
                                if_certainty = certainty })
           Nothing -> return Nothing
       
-- offered cops

p_offers :: Parser [PlayerName]
p_offers  =  (p_keyword "nottelling:" >> return [])
         <|> (do p_keyword "ofc\\"
                 offers <- many (p_keyword "ofc:" >> p_name <?> "offer")
                 p_keyword "ofc/"
                 return offers)
           
-- wrapper for messages from other players --

p_froms :: [PlayerName] -> Parser msg_t -> Parser [(PlayerName, msg_t)]
p_froms cops p_msg =
    do p_keyword "from\\"
       msg_map <- many (p_from p_msg)
       p_keyword "from/"
       --let msgs = map (fromJust . ((flip lookup) msg_map)) cops
       --return msgs
       return msg_map

p_from :: Parser msg_t -> Parser (PlayerName, msg_t)
p_from p_msg =
    do p_keyword "from:"
       source <- p_name <?> "message sender"
       msg <- p_msg <?> "message body"
       return (source, msg)


-- world update --

p_world_update :: NameToLoc -> [PlayerName] -> PlayerName -> Parser WorldUpdate
p_world_update name_to_loc cops robber =
    do p_keyword "wor\\"
       epoch <- (p_keyword "wor:" >> p_number) <?> "epoch"
       loot <- (p_keyword "rbd:" >> p_number) <?> "loot"
       dirty_cops <- p_dirty_cops <?> "dirty cops"
       slave_cops <- p_slave_cops <?> "controlled cops"
       false_accusations <- p_false_accusations <?> "false accusations"
       banks <- (p_banks name_to_loc) <?> "banks"
       evidence <- (p_evidence name_to_loc) <?> "evidence"
       smell <- (p_keyword "smell:" >> p_number) <?> "smell"
       players <- p_players name_to_loc
       p_keyword "wor/"
       let cop_states = map (fromJust . ((flip lookup) players)) cops
           robber_state = lookup robber players
       return (WorldUpdate { wu_epoch = epoch, wu_loot = loot,
                             wu_dirty_cops = dirty_cops,
                             wu_slave_cops = slave_cops,
                             wu_false_accusations = false_accusations,
                             wu_bank_values = banks,
                             wu_smell = if smell == 0 then Nothing
                                                      else Just smell,
                             wu_evidence = evidence,
                             wu_cop_states = cop_states,
                             wu_robber_state = robber_state })

p_banks :: NameToLoc -> Parser [(Loc, Int)]
p_banks name_to_loc =
    do p_keyword "bv\\"
       banks <- many (p_bank name_to_loc)
       p_keyword "bv/"
       return banks

p_bank :: NameToLoc -> Parser (Loc, Int)
p_bank name_to_loc =
    do p_keyword "bv:"
       loc_name <- p_name <?> "bank location"
       val <- p_number <?> "bank value"
       case name_to_loc loc_name of
           Just loc -> return (loc, val)
           Nothing -> error ("ERROR: bad bank location name: " ++ loc_name)

p_evidence :: NameToLoc -> Parser [(Loc, Int)]
p_evidence name_to_loc =
    do p_keyword "ev\\"
       evidence <- many (p_evidence_piece name_to_loc)
       p_keyword "ev/"
       return evidence

p_evidence_piece :: NameToLoc -> Parser (Loc, Int)
p_evidence_piece name_to_loc =
    do p_keyword "ev:"
       loc_name <- p_name <?> "evidence location"
       world <- p_number <?> "evidence world #"
       case name_to_loc loc_name of
           Just loc -> return (loc, world)
           Nothing -> error ("ERROR: bad evidence location name: " ++ loc_name)

p_players :: NameToLoc -> Parser [(PlayerName, PlayerState)]
p_players name_to_loc =
    do p_keyword "pl\\"
       players <- many (p_player name_to_loc)
       p_keyword "pl/"
       return players

p_player :: NameToLoc -> Parser (PlayerName, PlayerState)
p_player name_to_loc =
    do p_keyword "pl:"
       name <- p_name <?> "player name"
       loc_name <- p_name <?> "player location"
       job <- p_player_type <?> "player type"
       case name_to_loc loc_name of
           Just loc ->
               return (name, PlayerState { ps_loc = loc, ps_type = job })
           Nothing -> error ("ERROR: bad player location name: " ++ loc_name)

p_player_type :: Parser PlayerType
p_player_type  =  (p_keyword "cop-foot" >> return PT_cop_foot)
              <|> (p_keyword "cop-car" >> return PT_cop_car)
              <|> (p_keyword "robber" >> return PT_robber)

p_dirty_cops :: Parser [PlayerName]
p_dirty_cops =
    do p_keyword "dc\\"
       dirty_cops <- many (p_keyword "dc:" >> p_name <?> "dirty cop")
       p_keyword "dc/"
       return dirty_cops

p_slave_cops :: Parser [(PlayerName, PlayerName)]
p_slave_cops =
    do p_keyword "sc\\"
       master_slave_pairs <- many p_slave_cop
       p_keyword "sc/"
       return master_slave_pairs

p_slave_cop :: Parser (PlayerName, PlayerName)
p_slave_cop =
    do p_keyword "sc:"
       master <- p_name <?> "controller bot"
       slave <- p_name <?> "controlled bot"
       return (master, slave)

p_false_accusations :: Parser [(PlayerName, PlayerName)]
p_false_accusations =
    do p_keyword "fac\\"
       accusations <- many p_false_accusation
       p_keyword "fac/"
       return accusations
       
p_false_accusation :: Parser (PlayerName, PlayerName)
p_false_accusation =
    do p_keyword "fac:"
       accuser <- p_name <?> "accuser bot"
       victim <- p_name <?> "victim bot"
       return (accuser, victim)


-- world state --

p_world_skeleton ::
    Parser (WorldSkeleton, NameToLoc, LocToName, [PlayerName], PlayerName)
p_world_skeleton =
    do p_keyword "wsk\\"
       player_name <- (p_keyword "name:" >> p_name) <?> "player name"
       robber <- (p_keyword "robber:" >> p_name) <?> "robber name"
       cops <- many ((p_keyword "cop:" >> p_name) <?> "cop name")
       locations <- p_locs <?> "locations"
       streets <- p_streets <?> "streets"
       p_keyword "wsk/"
       let all_loc_names = map loc_name locations
           all_locs = [0..length locations - 1]
           loc_to_name_map :: M.Map Loc LocName
           loc_to_name_map = M.fromList (zip all_locs all_loc_names)
           loc_to_name loc =
               case M.lookup loc loc_to_name_map of
               Just loc_name -> loc_name
               Nothing -> error ("ERROR: phantom location: " ++ show loc)
           name_to_loc_map :: M.Map LocName Loc
           name_to_loc_map = M.fromList (zip all_loc_names all_locs)
           name_to_loc loc_name =
               case M.lookup loc_name name_to_loc_map of
               Just loc -> Just loc
               Nothing -> Nothing
           parser_name_to_loc error_str loc_name =
               case name_to_loc loc_name of
               Just loc -> loc
               Nothing ->
                   error ("ERROR: bad " ++ error_str ++ " name: " ++ loc_name)
           hq_name =
               case [loc_name loc | loc <- locations, loc_tag loc == LT_hq] of
               [name] -> name
               [] -> error "ERROR: no headquarters"
               hqs -> error ("ERROR: mulitple headquarters: " ++ show hqs)
           hq = parser_name_to_loc "headquarters" hq_name
           bank_names =
               [loc_name loc | loc <- locations, loc_tag loc == LT_bank]
           banks = map (parser_name_to_loc "bank") bank_names
           robber_start_name =
               case [loc_name loc | loc <- locations,
                     loc_tag loc == LT_rob_start] of
               [name] -> name
               [] -> error "ERROR: no robber start point"
               rs -> error ("ERROR: mulitple robber start points: " ++ show rs)
           robber_start =
               parser_name_to_loc "robber start point" robber_start_name
           (map_player_state, ps_to_node, node_to_ps) =
               build_player_state_map locations streets hq_name name_to_loc
           ws = WorldSkeleton { ws_my_name = player_name,
                                ws_cops = cops,
                                ws_robber = robber,
                                ws_all_locs = all_locs,
                                ws_loc_to_name = loc_to_name,
                                ws_name_to_loc = name_to_loc,
                                ws_banks = banks,
                                ws_hq = hq,
                                ws_robber_start = robber_start,
                                ws_ps_to_node = ps_to_node,
                                ws_node_to_ps = node_to_ps,
                                ws_map_ps = map_player_state }
       return (ws, name_to_loc, loc_to_name, cops, robber)

-- locations --

p_loc_type :: Parser LocType
p_loc_type  =  (p_keyword "hq" >> return LT_hq)
           <|> (p_keyword "bank" >> return LT_bank)
           <|> (p_keyword "robber-start" >> return LT_rob_start)
           <|> (p_keyword "ordinary" >> return LT_ord)

p_loc :: Parser Location
p_loc = do p_keyword "nod:"
           name <- p_name <?> "location name"
           tag <- p_loc_type <?> "location tag"
           coord_x <- p_number <?> "x-coord"
           coord_y <- p_number <?> "y-coord"
           return (Location { loc_name = name, loc_tag = tag,
                              loc_coords = (coord_x, coord_y) })


p_locs :: Parser [Location]
p_locs = do p_keyword "nod\\"
            locs <- many p_loc <?> "locations"
            p_keyword "nod/"
            return locs

-- streets --

p_car_only :: Parser Bool
p_car_only  =  (p_keyword "car" >> return True)
              <|> (p_keyword "foot" >> return False)

p_street :: Parser Street
p_street = do p_keyword "edg:"
              start <- p_name <?> "street start location"
              end <- p_name <?> "street end location"
              car_only <- p_car_only <?> "street type"
              return (Street { st_start = start, st_end = end,
                               st_car_only = car_only })


p_streets :: Parser [Street]
p_streets = do p_keyword "edg\\"
               streets <- many p_street
               p_keyword "edg/"
               return streets

-- general --

p_name :: Parser String
p_name = identifier robber_tokenizer

p_keyword :: String -> Parser ()
p_keyword = reserved robber_tokenizer

p_number :: Parser Int
p_number =
    do negative <- option False (char '-' >> return True)
       natural <- fmap fromInteger (natural robber_tokenizer) <?> "number"
       return (if negative then -natural else natural)

p_or_game_over :: Parser some_t -> Parser (Maybe some_t)
p_or_game_over parser  =  (p_keyword "game-over" >> return Nothing)
                      <|> (p_keyword "accused:" >> return Nothing) 
                      <|> (parser >>= return . Just)
--- tokenizer ---

robber_language_def :: LanguageDef parse_state
robber_language_def =
     LanguageDef { commentStart = "/*",
                   commentEnd = "*/",
                   commentLine = ";",
                   nestedComments = True,
                   identStart = letter <|> digit <|> oneOf "-_#()",
                   identLetter = letter <|> digit <|> oneOf "-_#()",
                   opStart = fail "no operators",
                   opLetter = fail "no operators",
                   reservedNames = [-- player type
                                    "cop-foot", "cop-car", "robber",
                                    -- edge type
                                    "car", "foot",
                                    -- node tag
                                    "hq", "bank", "robber-start", "ordinary",
                                    -- keywords
                                    "wsk\\", "wsk/", "nod\\", "nod/",
                                    "edg\\", "edg/",
                                    "name:", "robber:", "cop:",
                                    "nod:", "edg:"],
                   reservedOpNames = ["-"],
                   caseSensitive = False }

robber_tokenizer :: TokenParser parse_state
robber_tokenizer = makeTokenParser robber_language_def

