module Protocol where

import System
import IO
import Monad
import Text.ParserCombinators.Parsec
import Control.Concurrent
import Control.Concurrent.MVar
import System.Posix.Signals
import CPUTime
import Numeric

import State
import Parser
import Dbg

timeout_in_s :: Int
timeout_in_s = 4

enable_watchdog :: Bool
enable_watchdog = True

run_step :: (MVar (Maybe some_t) -> IO ()) -> MVar (Maybe some_t) -> Dbg ->
            String -> IO some_t
run_step step var dbg stage_name =
    do _ <- if enable_watchdog then scheduleAlarm timeout_in_s else return 0
       timeout_var <- newEmptyMVar
       let alarm = CatchOnce (putMVar timeout_var True)
       old_alarm <- installHandler realTimeAlarm alarm Nothing
       time_start <- getCPUTime
       putMVar var Nothing
       step_thread_id <- forkIO (step var >> putMVar timeout_var False)
       timed_out <- takeMVar timeout_var
       old_alarm <- installHandler realTimeAlarm alarm Nothing
       _ <- installHandler realTimeAlarm old_alarm Nothing
       _ <- if enable_watchdog then scheduleAlarm 0 else return 0
       when timed_out (killThread step_thread_id)
       time_end <- getCPUTime
       let time_diff =
             " (" ++ showFFloat (Just 4)
                     ((fromInteger (time_end - time_start))/1.0e12) " sec)"
       dbgPutStrLn dbg ("STATS: " ++ dbg_head dbg ++ stage_name ++
                        if timed_out then " (timed out)" else time_diff)
       take_mvar_or_die var stage_name

take_mvar_or_die :: MVar (Maybe some_t) -> String -> IO (some_t)
take_mvar_or_die var stage_name =
    tryTakeMVar var >>= \contents ->
    case contents of
    Nothing -> do hPutStrLn stderr
                      ("ERROR: stage " ++ show stage_name ++
                       " had an empty MVar -- that should not happen!")
                  exitWith (ExitFailure 3)
    Just maybe_result ->
        case maybe_result of
            Nothing -> do hPutStrLn stderr
                              ("ERROR: stage " ++ show stage_name ++
                               " did not compute a result")
                          exitWith (ExitFailure 3)
            Just result -> return result

send_reg :: String -> PlayerType -> IO ()
send_reg name player_type =
    do hSetBuffering stdin LineBuffering
       hSetBuffering stdout LineBuffering
       putStr "reg: "
       putStr name
       putChar ' '
       putStrLn (message (error "no loc_to_name") player_type)

send_move :: LocToName -> PlayerMove -> IO ()
send_move loc_to_name player_move =
    do putStr "mov: "
       putStrLn (message loc_to_name player_move)
       
send_moves :: LocToName -> [PlayerMove] -> IO ()
send_moves loc_to_name moves =
    do putStrLn "mov\\"
       mapM_ (send_move loc_to_name) moves
       putStrLn "mov/"
       
send_cop_move :: LocToName -> CopMove -> IO ()
send_cop_move loc_to_name move =
    do putStrLn "cmov\\"
       putStrLn (if cm_is_honest move then "straight-arrow:" else "turncoat:")
       send_moves loc_to_name (cm_moves move)
       putStrLn "acc\\"
       mapM_ (putStrLn . ("acc: " ++)) (cm_accusations move)
       putStrLn "acc/"
       putStrLn "cmov/"

send_robber_move :: LocToName -> RobberMove -> IO ()
send_robber_move loc_to_name move =
    do putStrLn "rmov\\"
       send_move loc_to_name (rm_move move)
       putStrLn (message loc_to_name (rm_bribe move))
       putStrLn "rmov/"

send_informs :: LocToName -> [Inform] -> IO ()
send_informs loc_to_name infs =
    do putStrLn "inf\\"
       mapM_ (putStrLn . ("inf: " ++) . (message loc_to_name)) infs
       putStrLn "inf/"
      
send_plans :: LocToName -> [Plan] -> IO ()
send_plans loc_to_name plans =
    do putStrLn "plan\\"
       mapM_ (putStrLn . ("plan: " ++) . (message loc_to_name)) plans
       putStrLn "plan/"

send_vote :: [PlayerName] -> IO ()
send_vote names =
    do putStrLn "vote\\"
       mapM_ (putStrLn . ("vote: " ++)) names
       putStrLn "vote/"

send_bribe :: Bool -> IO ()
send_bribe True = putStrLn "bribe:"
send_bribe False = putStrLn "nobribe:"

read_skeleton ::
    IO (WorldSkeleton, NameToLoc, LocToName, [PlayerName], PlayerName)
read_skeleton = parse_from_stdin p_world_skeleton (== "wsk/")

read_update :: NameToLoc -> [PlayerName] -> PlayerName -> IO WorldUpdate
read_update name_to_loc cops robber =
    parse_from_stdin (p_world_update name_to_loc cops robber) (== "wor/")

read_informs :: NameToLoc -> [PlayerName] -> IO [(PlayerName, [Maybe Inform])]
read_informs name_to_loc cops =
    parse_from_stdin (p_from_informs name_to_loc cops) (== "from/")

read_offers :: IO [PlayerName]
read_offers =
    parse_from_stdin p_offers (\ str -> (str == "ofc/") ||
                                        (str == "nottelling:"))

read_plans :: NameToLoc -> [PlayerName] -> IO [(PlayerName, [Maybe Plan])]
read_plans name_to_loc cops =
    parse_from_stdin (p_from_plans name_to_loc cops) (== "from/")

read_winner :: IO Winner
read_winner = parse_from_stdin p_winner (const True)

parse_from_stdin :: Parser some_t -> (String -> Bool) -> IO some_t
parse_from_stdin parser is_end =
    do lines <- read_lines_until is_end
       case runParser (p_or_game_over parser) () "<stdin>" (unlines lines) of
         Left err -> print err >> exitWith (ExitFailure 2)
         Right Nothing -> exitWith ExitSuccess
         Right (Just result) -> return result

read_lines_until :: (String -> Bool) -> IO [String]
read_lines_until is_end =
    do at_eof <- isEOF
       when at_eof (exitWith ExitSuccess)
       line <- getLine
       if is_end line
         then return [line]
         else read_lines_until is_end >>= return . (line :)

