module RobberHeuristic where

import qualified Data.Set as S
import Data.List

import State
import RobberAI
import AI
import Dbg
import LegalMoves
import PlotCourse
import GraphUtil
import RobberPlugins
import RobberPluginLog
import RobberMoveUtil

-- robber-heuristic goes for the richest bank, avoiding cops

-- ai config --

-- cost for distance-from-cop: 0, 1, ... in cash ("will risk it for $x") units
cost_per_dist :: [Int]
cost_per_dist = [100000, 10000, 1000, 100, 10] ++ repeat 1

max_radius :: Int
max_radius = length (takeWhile (> 1) cost_per_dist)

-- higher score = better chance of accepting plan
score :: (Int, Distance, Loc, [PlayerState]) -> Int
score inp@(cash, dist, locnam, path) =
    let final_score = cash - dist
--    in  trace ("SCORE: " ++ show final_score ++ " for " ++ locnam ++ " with $" ++ show cash ++ " at " ++ show dist) final_score
    in  final_score

-- ai implementation --

robber_heuristic =
    RobberAI { rai_name = "Heuristic",
               rai_init = robber_init,
               rai_inform = robber_inform,
               rai_plan = robber_plan,
               rai_vote = robber_vote,
               rai_bribe = robber_bribe,
               rai_move = robber_move }

data AIState = AIState {
    ais_dummy :: ()
}

robber_init :: RobberInitAI AIState
robber_init dbg plugins ws state_var =
    set_state state_var plugins
    AIState {
        ais_dummy = ()
    }

robber_inform :: RobberStepAI AIState WorldUpdate [Inform]
robber_inform dbg plugins state wu inform_var =
    set_informs dbg inform_var plugins state ([], "none")

robber_plan :: RobberStepAI state_t [(PlayerName, [Maybe Inform])] [Plan]
robber_plan dbg plugins state informs plan_var =
    set_plans dbg plan_var plugins state ([], "none")

robber_vote :: RobberStepAI state_t [(PlayerName, [Maybe Plan])] Vote
robber_vote dbg plugins state informs vote_var =
    do let ws = rpi_log_skeleton (rpi_log plugins)
           wu = rpi_log_update (rpi_log plugins)
           scoundrels = ws_my_name ws : wu_dirty_cops wu
       set_vote dbg vote_var plugins state (scoundrels, "none")

robber_bribe :: RobberStepAI state_t Winner Bool
robber_bribe dbg plugins state informs bribe_var =
    set_bribe dbg bribe_var plugins state (False, "no")

robber_move :: RobberStepAI AIState [PlayerName] RobberMove
robber_move dbg plugins state offered_cops result_var =
    do let identity_move_str = get_identity_move plugins
       set_move dbg result_var plugins state identity_move_str

       let wu = rpi_log_update (rpi_log plugins)
           Just here = wu_robber_state wu
           ws = rpi_log_skeleton (rpi_log plugins)
       let cops = wu_cop_states wu
           radii_maps = [(r, map_avoiding_cops ws cops r) | r <- [0..]]
       let plans_sets =
               [(care_radius,
                 [(val, cost, loc, path)
                  | (loc,val) <- wu_bank_values wu,
                    let (path,cost) = plot_course_with_map ws pmap here loc])
                 | (care_radius,pmap) <- take max_radius radii_maps]
       mapM_ (robber_move_path_set dbg plugins state result_var)
             plans_sets

robber_move_path_set dbg plugins state result_var (radius,plans) =
    do -- dbgPutStrLn dbg $ "RH: radius " ++ show radius
       let ((scr, path, bank) : _) = sort [(-(score plan), path, loc)
                                         | plan@(val, len, loc, path) <- plans]
           move = case path of
                  (here : next : _) -> next
                  [here] -> here
                  [] -> error "robber_move_path_set: phantom bank"
       bank `seq` scr `seq` return ()
       -- dbgPutStrLn dbg $ "RH: score for " ++ show bank ++ " = " ++ show scr
       -- dbgPutStrLn dbg $ "RH: move towards " ++ show bank ++ " = " ++ show move
       -- dbgPutStrLn dbg $ "RH: path: " ++ show path
       set_move dbg result_var plugins state
           (state_to_robber_move plugins move, "radius " ++ show radius)

-- robber-nodes at various distances from cops
nodesets_near_cops ws = remove_dupls . combine_node_sets . map (nodesets_near_cop ws)

-- robber-nodes at various distances from each cop
nodesets_near_cop ws cop_ps =
    let robbify ps = ps { ps_type = PT_robber }
        ps_to_r_node = ws_ps_to_node ws . robbify
        nss = remove_dupls $ map S.fromList $
            [ps_to_r_node cop_ps] :
            map (map ps_to_r_node . S.toList .
                 get_states_at_distance ws cop_ps) [1..]
    in  nss

map_avoiding_cops ws cops care_radius =
    let nss = take (care_radius + 1) (nodesets_near_cops ws cops)
        cost_dsts = concat [[(n,cst) | n <- S.toList ns]
                            | (ns,cst) <- zip nss cost_per_dist]
        mk_msg (n, cst) = "RH: costs: $" ++ show cst ++
                           " for " ++ show ((ws_node_to_ps ws) n,n) ++ "\n"
        new_map = repl_edges_to (ws_map_ps ws) cost_dsts
        msg1 = concatMap mk_msg cost_dsts
        msg2 = "RH: map:\n" ++ show new_map
--    in  trace msg1 $ trace msg2 $ new_map
    in  new_map

-- remove elements in earlier sets from later sets
remove_dupls :: (Ord elt_t) => [S.Set elt_t] -> [S.Set elt_t]
remove_dupls sets = remove_dupls' sets S.emptySet 

remove_dupls' :: (Ord elt_t) => [S.Set elt_t] -> S.Set elt_t -> [S.Set elt_t]
remove_dupls' [] _ = []
remove_dupls' (s:ss) seen = (s `S.minusSet` seen) : remove_dupls' ss (seen `S.union` s)

combine_node_sets :: (Ord elt_t) => [[S.Set elt_t]] -> [S.Set elt_t]
combine_node_sets setss =
    S.unions (map head setss) : combine_node_sets (map tail setss)

