module RobberJanosik where

import qualified Data.Set as S
import Data.List
import Data.Maybe
import IO
import Debug.Trace
import Monad
import Data.Graph.Inductive

import State
import RobberAI
import AI
import Dbg
import LegalMoves
import RandomUtil
import PlotCourse
import GraphUtil
import RobberPlugins
import RobberMoveUtil
import RobberPluginLog

-- robber-janosik goes for the richest bank, avoiding cops, and then hides
-- for a while

-- ai config --

-- cost for distance-from-cop: 0, 1, ... in cash ("will risk it for $x") units
cost_per_dist :: [Int]
cost_per_dist = [100000, 10000, 5000, 1000, 500] ++ repeat 1

max_radius :: Int
max_radius = length (takeWhile (> 1) cost_per_dist)

hideout_safety = 5
num_hideouts = 10
--max_heat = 10
--max_heat = 0

-- higher score = better chance of accepting plan
score :: (Int, Distance, Loc, [PlayerState]) -> Int
score inp@(cash, dist, locnam, path) =
    let final_score | cash == 0 = -1000000
                    | otherwise = cash - dist
--    in  trace ("SCORE: " ++ show final_score ++ " for " ++ locnam ++ " with $" ++ show cash ++ " at " ++ show dist) final_score
    in  final_score

-- ai implementation --

robber_janosik max_heat =
    RobberAI { rai_name = "Janosik",
               rai_init = robber_init max_heat,
               rai_inform = robber_inform,
               rai_plan = robber_plan,
               rai_vote = robber_vote,
               rai_bribe = robber_bribe,
               rai_move = robber_move }

data AIState = AIState {
    ais_max_heat :: Int,
    ais_heat :: Int,
    ais_hideouts :: [Loc],
    ais_banks :: S.Set Node,
    ais_map_no_banks :: Gr PlayerState Int,
    ais_bad_hideouts_locset :: S.Set Loc
}

bad_hideouts =
    ["greenwood-culdesac-#1", "greenwood-culdesac-#2",
     "quad-loop-#1", "quad-loop-#2", "quad-loop-#3",
     "museum-loop-#1", "museum-loop-#2", "museum-loop-#3",
     "museum-loop-#4", "museum-and-columbia",
     "science-and-columbia", "science-and-lake-shore",
     "ingleside-culdesac", "maryland-culdesac",
     "ellis-culdesac", "kimbark-culdesac", "52-culdesac",
     "52pl-and-harper(west)", "52pl-and-harper(east)",
     "south-shore-culdesac", "everett-culdesac-#1",
     "everett-culdesac-#2", "54pl-culdesac-#1",
     "54pl-culdesac-#2", "kenwood-culdesac-#1",
     "kenwood-culdesac-#2", "55-culdesac"
    ]


robber_init :: Int -> RobberInitAI AIState
robber_init max_heat dbg plugins ws state_var =
    set_state state_var plugins
    AIState {
        ais_max_heat = max_heat,
        ais_heat = max_heat,
        ais_hideouts = [],
        ais_banks = S.fromList [(ws_ps_to_node ws)
                                (PlayerState { ps_loc = bank,
                                              ps_type = PT_robber })
                                | bank <- ws_banks ws],
        ais_map_no_banks =
            repl_edges_to (ws_map_ps ws) [(n, 10000) | n <- bank_nodes ws],
        ais_bad_hideouts_locset =
            S.fromList (map (fromJust . (ws_name_to_loc ws)) bad_hideouts)
    }

robber_inform :: RobberStepAI AIState WorldUpdate [Inform]
robber_inform dbg plugins state wu inform_var =
    do let cops = wu_cop_states wu
       hideouts <- update_hideouts plugins state cops
       let ws = rpi_log_skeleton (rpi_log plugins)
           show_hide l = "\n        " ++ (ws_loc_to_name ws) l
--           inform_msg = "hideouts:" ++ concatMap show_hide hideouts
           inform_msg = "inform"
           state' = state { ais_hideouts = hideouts }
       set_informs dbg inform_var plugins state' ([], inform_msg)

robber_plan :: RobberStepAI state_t [(PlayerName, [Maybe Inform])] [Plan]
robber_plan dbg plugins state informs plan_var =
    set_plans dbg plan_var plugins state ([], "none")

robber_vote :: RobberStepAI state_t [(PlayerName, [Maybe Plan])] Vote
robber_vote dbg plugins state informs vote_var =
    do let ws = rpi_log_skeleton (rpi_log plugins)
           wu = rpi_log_update (rpi_log plugins)
           scoundrels = ws_my_name ws : wu_dirty_cops wu
       set_vote dbg vote_var plugins state (scoundrels, "dumb")

robber_bribe :: RobberStepAI state_t Winner Bool
robber_bribe dbg plugins state informs bribe_var =
    set_bribe dbg bribe_var plugins state (False, "no")

robber_move :: RobberStepAI AIState [PlayerName] RobberMove
robber_move dbg plugins state offered_cops result_var =
    do let identity_move_str = get_identity_move plugins
       set_move dbg result_var plugins state identity_move_str

       let wu = rpi_log_update (rpi_log plugins)
           ws = rpi_log_skeleton (rpi_log plugins)
           cops = wu_cop_states wu
           Just here = wu_robber_state wu
           old_heat = ais_heat state
           hot_nodeset = S.unions $
                         ais_banks state : take 2 (nodesets_near_cops ws cops)
           heat | (ws_ps_to_node ws) here `S.member` hot_nodeset =
                    ais_max_heat state
                | otherwise = old_heat
           heat' | heat > 0 = heat - 1
                 | otherwise = heat
           state' = state { ais_heat = heat' }
       dbgPutStrLn dbg ("HEAT: " ++ show heat)

       default_move <-
          if (heat > 0)
             then robber_move_hide dbg plugins state' wu result_var
             else robber_move_rob dbg plugins state' wu result_var

       search_moves <- get_search_moves plugins default_move
       mapM_ (set_move dbg result_var plugins state') search_moves

       return ()

-- if the heat is on, hide
--robber_move_hide :: RobberStepAI AIState WorldUpdate PlayerState
robber_move_hide dbg plugins state wu result_var =
    do let Just here = wu_robber_state wu
           ws = rpi_log_skeleton (rpi_log plugins)
           hideouts = ais_hideouts state
           cops = wu_cop_states wu
           radii_maps = [(r, map_avoiding_cops_banks state ws cops r) | r <- [0..]]
           best_hide_path (radius, hps) =
               let ((_, path, final):_) = sort hps in (radius, path, final)
           plans_sets = [(care_radius, [(cost, path, n) | n <- hideouts, let (path, cost) = plot_course_with_map ws pmap here n])
                      | (care_radius, pmap) <- take max_radius radii_maps]
           move_strs = map (robber_move_path_set
                            dbg plugins state result_var "hide" .
                            best_hide_path) plans_sets
       mapM_ (set_move dbg result_var plugins state) move_strs
       return (fst (last move_strs))

update_hideouts :: RobberPlugins -> AIState -> [PlayerState] -> IO [Loc]
update_hideouts plugins state cops =
    do let ws = rpi_log_skeleton (rpi_log plugins)
           hot_nodeset = S.unions $ take hideout_safety
                         (ais_banks state :
                          nodesets_near_cops ws cops)
           node_to_loc n =
               let PlayerState { ps_loc = loc } = (ws_node_to_ps ws) n
               in  loc
           hot_locset = S.map node_to_loc hot_nodeset
                        `S.union`
                        ais_bad_hideouts_locset state
           safe_hides = [h | h <- ais_hideouts state,
                             not (h `S.member` hot_locset)]
           n_more_hides = num_hideouts - length safe_hides
           cold_set = S.fromList (ws_all_locs ws) `S.difference` hot_locset
           n_cold_set = S.size cold_set
           pick_hide (places, new_hides) _ =
               do hide <- random_list_elt places
                  return (delete hide places, hide : new_hides)
       (_, new_hides) <- foldM pick_hide (S.toList cold_set, safe_hides)
                               [0 .. (num_hideouts - length safe_hides) - 1]
       return new_hides


-- when the heat is not on, go rob something
--robber_move_rob :: RobberStepAI AIState WorldUpdate PlayerState
robber_move_rob dbg plugins state wu result_var =
    do let Just here = wu_robber_state wu
           ws = rpi_log_skeleton (rpi_log plugins)
           cops = wu_cop_states wu
           radii_maps = [(r, map_avoiding_cops state ws cops r) | r <- [0..]]
           best_bank_path (radius, bps) =
               let ((_, path, final):_) = sort [(-(score plan), path, loc)
                                        | plan@(val, len, loc, path) <- bps]
               in  (radius, path, final)
           Just boonies = (ws_name_to_loc ws) "museum-loop-#3"
           plans_sets =
               [(care_radius,
                 [(val, cost + ret_cost, loc, path)
                  | (loc,val) <- wu_bank_values wu,
                    let (path,cost) = plot_course_with_map ws pmap here loc,
                    let loc_st = PlayerState { ps_type = PT_robber, ps_loc = loc },
                    let (ret_path,ret_cost) =
                            plot_course_with_map ws pmap loc_st boonies])
                 | (care_radius,pmap) <- take max_radius radii_maps]
           move_strs = map (robber_move_path_set
                            dbg plugins state result_var "bank" .
                            best_bank_path) plans_sets
       mapM_ (set_move dbg result_var plugins state) move_strs
       return (fst (last move_strs))

robber_move_path_set dbg plugins state result_var str (radius,path,final) =
    let move = case path of
               (here : next : _) -> next
               [here] -> here
               [] -> error "robber_janosik_move_path_set: phantom bank/hideout"
        ws = rpi_log_skeleton (rpi_log plugins)
        target = (ws_loc_to_name ws) (ps_loc (last path))
        path_str = concat ["\n        -> " ++ (ws_loc_to_name ws) (ps_loc l)
                           | l <- path]
    in  (state_to_robber_move plugins move,
         str ++ ": radius " ++ show radius ++ " -> " ++
         (ws_loc_to_name ws) final)

-- robber-nodes at various distances from cops
nodesets_near_cops ws = remove_dupls . combine_node_sets . map (nodesets_near_cop ws)

-- robber-nodes at various distances from each cop
nodesets_near_cop ws cop_ps =
    let robbify ps = ps { ps_type = PT_robber }
        ps_to_r_node = ws_ps_to_node ws . robbify
        nss = remove_dupls $ map S.fromList $
            [ps_to_r_node cop_ps] :
            map (map ps_to_r_node . S.toList .
                 get_states_at_distance ws cop_ps) [1..]
    in  nss

map_avoiding_cops state ws cops care_radius =
    map_avoiding_cops_from_map ws (ws_map_ps ws) cops care_radius

map_avoiding_cops_banks state ws cops care_radius =
    map_avoiding_cops_from_map ws (ais_map_no_banks state) cops care_radius

map_avoiding_cops_from_map ws orig_map cops care_radius =
    let nss = take (care_radius + 1) (nodesets_near_cops ws cops)
        cost_dsts = concat [[(n,cst) | n <- S.toList ns]
                            | (ns,cst) <- zip nss cost_per_dist]
        mk_msg (n, cst) = "RH: costs: $" ++ show cst ++
                           " for " ++ show ((ws_node_to_ps ws) n,n) ++ "\n"
        new_map = repl_edges_to orig_map cost_dsts
        msg1 = concatMap mk_msg cost_dsts
        msg2 = "RH: map:\n" ++ show new_map
--    in  trace msg1 $ trace msg2 $ new_map
    in  new_map

bank_nodes ws =
    let robbify loc = PlayerState { ps_type = PT_robber, ps_loc = loc }
    in  map ((ws_ps_to_node ws) . robbify) (ws_banks ws)

-- remove elements in earlier sets from later sets
remove_dupls :: (Ord elt_t) => [S.Set elt_t] -> [S.Set elt_t]
remove_dupls sets = remove_dupls' sets S.emptySet 

remove_dupls' :: (Ord elt_t) => [S.Set elt_t] -> S.Set elt_t -> [S.Set elt_t]
remove_dupls' [] _ = []
remove_dupls' (s:ss) seen = (s `S.difference` seen) : remove_dupls' ss (seen `S.union` s)

combine_node_sets :: (Ord elt_t) => [[S.Set elt_t]] -> [S.Set elt_t]
combine_node_sets setss =
    S.unions (map head setss) : combine_node_sets (map tail setss)

