module RobberMoveUtil where

import List
import Maybe
import Random

import RobberPlugins
import State
import RobberPluginLog
import LegalMoves
import RandomUtil
import CatchSearch
import RobberPluginCache
import Minimax

state_to_player_move :: PlayerName -> PlayerState -> PlayerMove
state_to_player_move player_name player_state =
    PlayerMove { pm_name = player_name,
                 pm_state = player_state }

player_move_to_robber_move :: PlayerMove -> RobberMove
player_move_to_robber_move player_move =
    RobberMove { rm_move = player_move,
                 rm_bribe = BR_none }

state_to_robber_move :: RobberPlugins -> PlayerState -> RobberMove
state_to_robber_move plugins robber_state =
    let ws = rpi_log_skeleton (rpi_log plugins)
        my_name = ws_my_name ws
    in player_move_to_robber_move $
       state_to_player_move my_name robber_state

get_identity_move :: RobberPlugins -> (RobberMove, String)
get_identity_move plugins =
    let ws = rpi_log_skeleton (rpi_log plugins)
        wu = rpi_log_update (rpi_log plugins)
        Just current_ps = wu_robber_state wu
        move = state_to_robber_move plugins current_ps
    in  (move, "identity")

get_random_move :: RobberPlugins -> IO (RobberMove, String)
get_random_move plugins =
    do let ws = rpi_log_skeleton (rpi_log plugins)
           wu = rpi_log_update (rpi_log plugins)
           Just current_ps = wu_robber_state wu
           legal_states = legal_moves ws current_ps
       random_ps <- random_list_elt legal_states
       let move = state_to_robber_move plugins random_ps
       return (move, "random")

get_random_non_identity_move :: RobberPlugins -> IO (RobberMove, String)
get_random_non_identity_move plugins =
    do let ws = rpi_log_skeleton (rpi_log plugins)
           wu = rpi_log_update (rpi_log plugins)
           Just current_ps = wu_robber_state wu
           legal_states = legal_moves ws current_ps
           legal_non_identity_states = delete current_ps legal_states
           states_to_use =
               if null legal_non_identity_states
                 then legal_states
                 else legal_non_identity_states
       random_ps <- random_list_elt states_to_use
       let move = state_to_robber_move plugins random_ps
       return (move, "random non-identity")

get_search_moves :: RobberPlugins -> RobberMove -> IO [(RobberMove, String)]
get_search_moves plugins default_move =
    do let ws = rpi_log_skeleton (rpi_log plugins)
           wu = rpi_log_update (rpi_log plugins)
           ss = SearchState {
                    ss_epoch = wu_epoch wu,
                    ss_cop_states = map Just (wu_cop_states wu),
                    ss_robber_state = fromJust (wu_robber_state wu),
                    ss_last_smell_epoch = 0 }
           default_state = pm_state (rm_move (default_move))

           locs_up_to_distance_cached =
               rpi_cache_locs_up_to_distance (rpi_cache plugins)
           locs_up_to_distance_same_state_cached =
               rpi_cache_locs_up_to_distance_same_state (rpi_cache plugins)

           opts = make_minimax_opts ws
                  locs_up_to_distance_cached
                  locs_up_to_distance_same_state_cached
                  [Just (SM_RobberMove default_state)]
       minimax_g <- newStdGen
       let depths = drop 1 (iterate (mmo_next_depth opts) 0)
           scores_and_moves = minimax opts minimax_g ss
           moves = [(state_to_robber_move plugins move, str)
                    |
                    ((score, (SM_RobberMove move:_)), depth) <-
                        zip scores_and_moves depths,
                    let str = "search depth " ++ show depth ++
                              (if move == default_state
                                 then " (no override)"
                                 else " (OVERRIDE)") ++
                              " (score = " ++ show score ++ ")"]
       return moves
