module Search where

import Maybe

import qualified Data.Set as S

import Minimax
import State
import LegalMoves
import ListUtil

make_minimax_opts :: WorldSkeleton ->
                     LocsUpToDistanceFn ->
                     LocsUpToDistanceFn ->
                     [Maybe SearchMove] ->
                     MinimaxOpts SearchState SearchMove SearchScore
make_minimax_opts ws
                  locs_up_to_distance
                  locs_up_to_distance_same_state
                  default_moves =
    default_minimax_opts {
        mmo_moves = search_moves ws locs_up_to_distance,
        mmo_advance = search_advance locs_up_to_distance_same_state,
        mmo_eval = search_eval,
        mmo_want_max = search_want_max,
        mmo_next_depth = (+1),
        mmo_default_moves = default_moves
    }

data SearchState = SearchState {
    ss_epoch :: Epoch,
    ss_cop_states :: [Maybe PlayerState],
    ss_robber_state :: PlayerState,
    ss_last_smell_epoch :: Epoch
}

data SearchMove =
    RobberMove PlayerState | CopMove [Maybe PlayerState]
    deriving (Eq, Show)

data SearchScore =
    TimeIsUp |
    GameNotOver Int |
    RobberCaught Int
    deriving (Eq, Ord, Show)

search_moves :: WorldSkeleton -> LocsUpToDistanceFn -> Depth -> SearchState ->
                [SearchMove]
search_moves ws locs_up_to_distance depth ss
    | is_time_up ss = []
    | is_robber_caught ss = []
    | robber_to_move (ss_epoch ss) =
        map RobberMove (legal_moves ws (ss_robber_state ss))
    | cop_to_move (ss_epoch ss) =
        map CopMove (cartesian_product
                     (map (search_moves_cop ws
                                            locs_up_to_distance
                                            depth
                                            (ss_robber_state ss))
                              (ss_cop_states ss)))

search_moves_cop ::
    WorldSkeleton -> LocsUpToDistanceFn -> Depth -> PlayerState ->
    Maybe PlayerState -> [Maybe PlayerState]
search_moves_cop _ _ _ _ Nothing = [Nothing]
search_moves_cop ws locs_up_to_distance depth robber_state (Just cop_state) =
    let cop_move_depth = (depth + 1) `div` 2
        robber_depth = depth - cop_move_depth
        smell_dist = get_smell_dist (ps_type cop_state)
        cop_smell_depth = robber_depth + smell_dist
        cop_depth = max cop_move_depth cop_smell_depth
        robber_locs = (locs_up_to_distance robber_state) !! robber_depth
        all_moves = legal_moves ws cop_state
        moves = [Just move |
                 move <- all_moves,
                 let cop_locs = (locs_up_to_distance move) !! (cop_depth - 1),
                 not (S.null (cop_locs `S.intersection` robber_locs))]
    in  if length moves < length all_moves
          then (Nothing:moves)
          else moves


search_advance :: LocsUpToDistanceFn -> SearchState ->
                  SearchMove -> SearchState
search_advance locs_up_to_distance_same_state ss (RobberMove robber_state) =
    let ss' = ss { ss_epoch = ss_epoch ss + 1,
                   ss_robber_state = robber_state }
    in update_last_smell_epoch locs_up_to_distance_same_state ss'
search_advance _ ss (CopMove cop_states) =
    ss { ss_epoch = ss_epoch ss + 1,
         ss_cop_states = cop_states }

update_last_smell_epoch :: LocsUpToDistanceFn ->
                           SearchState -> SearchState
update_last_smell_epoch locs_up_to_distance_same_state ss =
    let cop_states = [fromJust maybe_cop_state |
                      maybe_cop_state <- ss_cop_states ss,
                      isJust maybe_cop_state]
        robber_state = ss_robber_state ss
        smells_list = map (cop_smells_robber
                           locs_up_to_distance_same_state
                           robber_state) cop_states
    in  if or smells_list
          then ss { ss_last_smell_epoch = (ss_epoch ss) }
          else ss

cop_smells_robber :: LocsUpToDistanceFn -> PlayerState -> PlayerState -> Bool
cop_smells_robber locs_up_to_distance_same_state robber_state cop_state =
    let smell_dist = get_smell_dist (ps_type cop_state)
        smell_locs = (locs_up_to_distance_same_state cop_state) !! smell_dist
        robber_loc = state_to_loc robber_state
    in  robber_loc `S.member` smell_locs

search_eval :: SearchState -> SearchScore
search_eval ss
    | is_robber_caught ss = RobberCaught (-(ss_epoch ss))
    | is_time_up ss = TimeIsUp
    | otherwise = GameNotOver (ss_last_smell_epoch ss)

search_want_max :: SearchState -> Bool
search_want_max ss = cop_to_move (ss_epoch ss)

is_time_up :: SearchState -> Bool
is_time_up ss =
    ss_epoch ss == 200

is_robber_caught :: SearchState -> Bool
is_robber_caught ss =
    let robber_loc = state_to_loc (ss_robber_state ss)
        cop_states = [fromJust maybe_cop_state |
                      maybe_cop_state <- ss_cop_states ss,
                      isJust maybe_cop_state]
        cop_locs = map state_to_loc cop_states
    in  robber_loc `elem` cop_locs
