module State where

import List
import qualified Data.Graph as G
import qualified Data.Map as M
import Data.Graph.Inductive
import Data.Graph.Inductive.Monad.IOArray
import Text.Show.Functions
import qualified Data.Set as S
import Data.Maybe

class Message some_t where
    message :: (Loc -> LocName) -> some_t -> String

data WorldSkeleton = WorldSkeleton {
    ws_my_name :: PlayerName,
    ws_cops :: [PlayerName],
    ws_robber :: PlayerName,
    ws_all_locs :: [Loc],
    ws_loc_to_name :: Loc -> LocName,
    ws_name_to_loc :: LocName -> Maybe Loc,
    ws_hq :: Loc,
    ws_robber_start :: Loc,
    ws_banks :: [Loc],
    ws_ps_to_node :: PlayerState -> Node,
    ws_node_to_ps :: Node -> PlayerState,
    ws_map_ps :: PlayerStateMap
} deriving (Show)


data WorldUpdate = WorldUpdate {
    wu_epoch :: Epoch,
    wu_loot :: Int,
    wu_dirty_cops :: [PlayerName],
    wu_slave_cops :: [(PlayerName, PlayerName)], -- (master, slave)
    wu_false_accusations :: [(PlayerName, PlayerName)], -- (accuser, victim)
    wu_bank_values :: [(Loc, Int)],
    wu_evidence :: [(Loc, Epoch)],
    wu_smell :: Maybe Distance,
    wu_cop_states :: [PlayerState], -- same order as ws_cops
    wu_robber_state :: Maybe PlayerState
} deriving (Show)

data Inform = Inform {
    if_who :: !PlayerName,
    if_loc :: !Loc,
    if_ptype :: !PlayerType,
    if_epoch :: !Epoch,
    if_certainty :: !Certainty
} deriving (Show)

instance Message Inform where
    message loc_to_name inf =
        if_who inf ++ " " ++ loc_to_name (if_loc inf) ++ " " ++
        message loc_to_name (if_ptype inf) ++ " " ++
        message loc_to_name (if_epoch inf) ++ " " ++
        message loc_to_name (if_certainty inf)

data Plan = Plan {
    pl_epoch :: !Epoch,
    pl_move :: !PlayerMove
} deriving (Show)

data CopMove = CopMove {
    cm_is_honest :: Bool,
    cm_accusations :: [PlayerName],
    cm_moves :: [PlayerMove]
} deriving (Eq, Show)

data RobberMove = RobberMove {
    rm_move :: PlayerMove,
    rm_bribe :: BribeResult
} deriving (Eq, Show)

instance Message Plan where
    message loc_to_name plan =
        pm_name (pl_move plan) ++ " " ++
        message loc_to_name (pm_state (pl_move plan)) ++ " " ++
        message loc_to_name (pl_epoch plan)

data LocType = LT_hq
             | LT_bank
             | LT_rob_start
             | LT_ord
    deriving (Eq, Show)

data Location = Location { loc_name :: LocName,
                           loc_tag :: LocType,
                           loc_coords :: (Int, Int) }
    deriving (Show)

data PlayerType = PT_cop_foot | PT_cop_car | PT_robber
    deriving (Eq, Ord)

instance Show PlayerType where
    show PT_cop_foot = "cop-foot"
    show PT_cop_car = "cop-car"
    show PT_robber = "robber"

instance Message PlayerType where
    message _ pt = show pt

data PlayerState = PlayerState {
    ps_loc :: !Loc,
    ps_type :: !PlayerType
} deriving (Eq, Ord)

data PlayerMove = PlayerMove {
    pm_name :: PlayerName,
    pm_state :: PlayerState
} deriving Eq

instance Show PlayerState where
    show (PlayerState { ps_loc = loc, ps_type = pt }) =
        show pt ++ "@" ++ show loc

instance Show PlayerMove where
    show (PlayerMove { pm_name = name, pm_state = state }) =
        show name ++ ":" ++ show state

        
type PlayerStateMap = Gr PlayerState Int -- # is travel cost

state_to_loc :: PlayerState -> Loc
state_to_loc player_state = ps_loc player_state

loc_to_state :: PlayerType -> Loc -> PlayerState
loc_to_state player_type loc =
    PlayerState { ps_loc = loc, ps_type = player_type }

get_smell_dist :: PlayerType -> Distance
get_smell_dist PT_cop_foot = 2
get_smell_dist PT_cop_car = 1

instance Message PlayerState where
    message loc_to_name pstate =
        loc_to_name (ps_loc pstate) ++ " " ++
        message loc_to_name (ps_type pstate)
        
instance Message PlayerMove where
    message loc_to_name move =
        message loc_to_name (pm_state move) ++ " " ++ pm_name move

type Vote = [PlayerName]
type Winner = Maybe PlayerName

type PlayerName = String
type LocName = String
type Epoch = Int
type Distance = Int
type Certainty = Int

instance Message Int where
    message _ n = show n

data BribeResult =
    BR_none | BR_push (PlayerName, Loc) | BR_choose [PlayerName]
    deriving (Eq, Show)

instance Message BribeResult where
    message loc_to_name BR_none = "nobribe:"
    message loc_to_name (BR_push (name, loc)) =
        "push: " ++ name ++ " " ++ message loc_to_name loc
    message loc_to_name (BR_choose cops) =
        "chc\\\n" ++
        concat ["chc: " ++ cop ++ "\n" | cop <- cops] ++
        "chc/"

robber_to_move epoch = epoch `mod` 2 == 0
cop_to_move epoch = epoch `mod` 2 == 1

data StreetType = ST_car | ST_foot
    deriving (Eq, Show)

data Street = Street { st_start :: LocName,
                       st_end :: LocName,
                       st_car_only :: Bool }
    deriving (Show)

build_player_state_map ::
    [Location] -> [Street] -> LocName -> (LocName -> Maybe Loc) ->
    (PlayerStateMap, PlayerState -> Node, Node -> PlayerState)
build_player_state_map locations streets hq name_to_loc =
    let all_states =
            [PlayerState { ps_loc = l, ps_type = t }
             | loc <- locations, let l = fromJust (name_to_loc (loc_name loc)),
               t <- [PT_robber, PT_cop_foot, PT_cop_car]]
        uniq_nodes :: [(Node, PlayerState)]
        uniq_nodes = zip [0..] all_states
        node_to_ps_map :: M.Map Node PlayerState
        node_to_ps_map = M.fromList uniq_nodes
        node_to_ps n = case M.lookup n node_to_ps_map of
                        Just ps -> ps
                        Nothing -> error ("ERROR: phantom node: " ++ show n)
        ps_to_node_map :: M.Map PlayerState Node
        ps_to_node_map = M.fromList (zip all_states [0..])
        ps_to_node ps = case M.lookup ps ps_to_node_map of
                       Just n -> n
                       Nothing -> error ("ERROR: phantom state: " ++ show ps)

        name_to_loc' loc_name =
            case name_to_loc loc_name of
            Just loc -> loc
            Nothing -> error ("ERROR: bad location name: " ++ loc_name)
        mk_edge start_pt end_pt start_loc end_loc =
            (PlayerState { ps_type = start_pt,
                           ps_loc = name_to_loc' start_loc },
             PlayerState { ps_type = end_pt,
                           ps_loc = name_to_loc' end_loc })
        mk_simple_edge pt start_loc end_loc = mk_edge pt pt start_loc end_loc

        street_to_edge st | st_car_only st =
           -- cops can drive on any car-only street in the right direction
           [mk_simple_edge PT_cop_car (st_start st) (st_end st)]
        street_to_edge st | otherwise =
           -- cops can drive on any normal street in the right direction
           [mk_simple_edge PT_cop_car (st_start st) (st_end st),
           -- cops can walk on any normal street in any direction
            mk_simple_edge PT_cop_foot (st_start st) (st_end st),
            mk_simple_edge PT_cop_foot (st_end st) (st_start st),
           -- robbers can walk on any normal street in any direction
            mk_simple_edge PT_robber (st_start st) (st_end st),
            mk_simple_edge PT_robber (st_end st) (st_start st)]
        mk_hq_edges st | st_start st == hq && st_car_only st =
           -- cops can take a car from hq to end of any car-only street
           [mk_edge PT_cop_foot PT_cop_car hq (st_end st)]
        mk_hq_edges st | st_start st == hq && not (st_car_only st) =
           -- cops can take a car from hq to end of any normal street
           [mk_edge PT_cop_foot PT_cop_car hq (st_end st),
           -- cops can leave a car at hq and walk to end of any normal street
            mk_edge PT_cop_car PT_cop_foot hq (st_end st)]
        mk_hq_edges st | st_end st == hq && not (st_car_only st) =
           -- cops can leave a car at hq and walk to start of any normal street
           [mk_edge PT_cop_car PT_cop_foot hq (st_start st)]
        mk_hq_edges st | otherwise = []
        stationary_hq_edges =
           -- cops can leave a car at hq and stay there on foot
           [mk_edge PT_cop_car PT_cop_foot hq hq,
           -- cops can get into a car at hq and stay there
            mk_edge PT_cop_foot PT_cop_car hq hq]

        --no_op_edges = [(ps, ps) | ps <- all_states]
        -- It turns out the graph library does not handle edges
        -- from a node to itself very vell.  For instance, at least
        -- in some cases suc will return the input node but pre
        -- will not.  Now moves from a state to itself are handled
        -- in legal_moves and legal_anti_moves directly.
        no_op_edges = []

        all_graph_edges = [(ps_to_node start, ps_to_node end, 1)
                           | (start, end) <- concat
                                [stationary_hq_edges, no_op_edges,
                                 concatMap mk_hq_edges streets,
                                 concatMap street_to_edge streets]]
        uniq_graph_edges = nub (sort all_graph_edges)
    in  (mkGraph uniq_nodes uniq_graph_edges, ps_to_node, node_to_ps)

type Loc = Int
type LocSet = S.Set Loc

type PlayerStateSet = S.Set PlayerState

type ImpossibleList = [LocSet]

lsSize ls = S.size ls
lsMember x ls = S.member x ls

lsEmpty = S.empty
lsSingleton x = S.singleton x

lsFromList l = S.fromList l
lsToList ls = S.toList ls

lsDelete ls1 x = S.delete x ls1

lsUnion ls1 ls2 = S.union ls1 ls2
lsUnions lsList = S.unions lsList
lsDifference ls1 ls2 = S.difference ls1 ls2
lsIntersection ls1 ls2 = S.intersection ls1 ls2

type PlayerStateLocSetMap = PlayerState -> LocSet
