module Tracking where

import Data.Array
import qualified Data.Set as S
import Maybe

import State
import LegalMoves

type Tracking = Array Int PlayerStateSet

tracking_epoch_to_idx :: Epoch -> Int
tracking_epoch_to_idx epoch = (epoch + 1) `div` 2

tracking_init :: WorldSkeleton -> Tracking
tracking_init ws =
    let robber_start_loc = ws_robber_start ws
        robber_start_state = PlayerState { ps_loc = robber_start_loc,
                                           ps_type = PT_robber }
    in listArray (0, 0) [lsSingleton robber_start_state]

tracking_advance :: LegalMovesFn -> Tracking -> Tracking
tracking_advance legal_moves_fn tracking =
    let (_, high) = bounds tracking
        next_possible_states =
            possible_next_states legal_moves_fn (tracking ! high)
    in  listArray (0, high + 1) (elems tracking ++ [next_possible_states])

possible_next_states :: LegalMovesFn -> PlayerStateSet -> PlayerStateSet
possible_next_states legal_moves_fn states =
    S.fromList (concatMap legal_moves_fn (S.toList states))

possible_prev_states :: LegalMovesFn -> PlayerStateSet -> PlayerStateSet
possible_prev_states legal_anti_moves_fn states =
    S.fromList (concatMap legal_anti_moves_fn (S.toList states))

tracking_update_possible ::
    LegalMovesFn -> LegalMovesFn ->
    Epoch -> PlayerStateSet -> Tracking -> Tracking
tracking_update_possible = tracking_update True

tracking_update_impossible ::
    LegalMovesFn -> LegalMovesFn ->
    Epoch -> PlayerStateSet -> Tracking -> Tracking
tracking_update_impossible = tracking_update False

tracking_update ::
    Bool -> LegalMovesFn -> LegalMovesFn ->
    Epoch -> PlayerStateSet -> Tracking -> Tracking
tracking_update possible legal_moves_fn legal_anti_moves_fn
                epoch states tracking =
    let merge_states = (if possible then S.intersection else S.difference)
        epoch_idx = tracking_epoch_to_idx epoch
        b = bounds tracking
        tracking' = listArray b [update_impossible_states idx | idx <- range b]
        changed = listArray b [index_changed idx | idx <- range b]
        update_impossible_states idx
            | idx < epoch_idx =
                if changed ! (idx + 1)
                   then (tracking ! idx)
                        `S.intersection`
                        (possible_prev_states legal_anti_moves_fn
                         (tracking' ! (idx + 1)))
                   else (tracking ! idx)
            | idx == epoch_idx =
                (tracking ! idx)
                `merge_states`
                states
            | idx > epoch_idx =
                if changed ! (idx - 1)
                   then (tracking ! idx)
                        `S.intersection`
                        (possible_next_states legal_moves_fn
                         (tracking' ! (idx - 1)))
                   else (tracking ! idx)
        index_changed idx
            | idx < epoch_idx =
                changed ! (idx + 1) &&
                tracking ! idx /= tracking' ! idx
            | idx == epoch_idx = 
                tracking ! idx /= tracking' ! idx
            | idx > epoch_idx =
                changed ! (idx - 1) &&
                tracking ! idx /= tracking' ! idx
    in  tracking'

-- In EvidenceTrackingItem, before calling
-- evidence_tracking_apply_robber_tracking the PlayerStateSet
-- represents the places where the evidence *cannot* be (because a cop
-- has been there), but after calling that long-named function it
-- represents the places where the evidence *can* be.  Just to confuse
-- you.

type EvidenceTrackingItem = (Epoch, PlayerStateSet)
type EvidenceTracking = [EvidenceTrackingItem]

evidence_tracking_init :: EvidenceTracking
evidence_tracking_init = []

evidence_tracking_advance :: Epoch -> EvidenceTracking -> EvidenceTracking
evidence_tracking_advance epoch et =
    let -- Remove expired evidence
        -- (changed 24 to 26 to match server behavior)
        et' = [item | item <- et, fst item /= epoch - 26]
        -- Add new evidence
        et'' = if (epoch `mod` 8 == 0) && (epoch /= 0)
                  then et' ++ [(epoch, S.empty)]
                  else et'
    in  et''

evidence_tracking_update ::
    WorldSkeleton -> WorldUpdate -> EvidenceTracking -> EvidenceTracking
evidence_tracking_update ws wu et =
    let cop_states = wu_cop_states wu
        robber_states =
            map ((loc_to_state PT_robber) . state_to_loc) cop_states
        robber_states_set = S.fromList robber_states
    in  map (evidence_tracking_update' robber_states_set) et

evidence_tracking_update' ::
    PlayerStateSet -> EvidenceTrackingItem -> EvidenceTrackingItem
evidence_tracking_update' new_states (epoch, item_states) =
    (epoch, item_states `S.union` new_states)

evidence_tracking_apply_robber_tracking ::
    EvidenceTracking -> Tracking -> EvidenceTracking
evidence_tracking_apply_robber_tracking et tracking =
    let et' = map (evidence_tracking_apply_robber_tracking' tracking) et
        et'' = [item | item <- et', not (S.null (snd item))]
    in  et''

evidence_tracking_apply_robber_tracking' ::
    Tracking -> EvidenceTrackingItem -> EvidenceTrackingItem
evidence_tracking_apply_robber_tracking' tracking (epoch, ev_set) =
    let tracking_idx = tracking_epoch_to_idx epoch
        tracking_set = tracking ! tracking_idx
    in  (epoch, tracking_set `S.difference` ev_set)
